from relsad.network.components import (
    Battery,
    Bus,
    CircuitBreaker,
    Disconnector,
    Line,
    ManualMainController,
    MicrogridMode,
    Production,
)
from relsad.network.systems import (
    Distribution,
    Microgrid,
    PowerSystem,
    Transmission,
)
from relsad.StatDist import NormalParameters, StatDist, StatDistType
from relsad.Table import Table
from relsad.Time import Time, TimeUnit
from relsad.visualization.plotting import plot_topology


def initialize_network(
    fail_rate_trafo: float = 0.007,
    fail_rate_line: float = 0.7,
    outage_time_trafo: Time = Time(8, TimeUnit.HOUR),
    microgrid_mode: MicrogridMode = MicrogridMode.SURVIVAL,
):

    line_stat_dist = StatDist(
        stat_dist_type=StatDistType.TRUNCNORMAL,
        parameters=NormalParameters(
            loc=1.25,
            scale=1,
            min_val=0.5,
            max_val=2,
        ),
    )

    C1 = ManualMainController(
        name="C1",
        sectioning_time=Time(1, TimeUnit.HOUR),
    )

    ps = PowerSystem(C1)

    B1 = Bus(
        name="B1",
        coordinate=[0, 0],
        fail_rate_per_year=0,
    )
    B2 = Bus(
        name="B2",
        coordinate=[1, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B3 = Bus(
        name="B3",
        coordinate=[2, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B4 = Bus(
        name="B4",
        coordinate=[3, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B5 = Bus(
        name="B5",
        coordinate=[4, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B6 = Bus(
        name="B6",
        coordinate=[5, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B7 = Bus(
        name="B7",
        coordinate=[6, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B8 = Bus(
        name="B8",
        coordinate=[7, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B9 = Bus(
        name="B9",
        coordinate=[8, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B10 = Bus(
        name="B10",
        coordinate=[9, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B11 = Bus(
        name="B11",
        coordinate=[10, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B12 = Bus(
        name="B12",
        coordinate=[11, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B13 = Bus(
        name="B13",
        coordinate=[12, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B14 = Bus(
        name="B14",
        coordinate=[13, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B15 = Bus(
        name="B15",
        coordinate=[14, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B16 = Bus(
        name="B16",
        coordinate=[15, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B17 = Bus(
        name="B17",
        coordinate=[16, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B18 = Bus(
        name="B18",
        coordinate=[17, 0.5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B19 = Bus(
        name="B19",
        coordinate=[18, 0.5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B20 = Bus(
        name="B20",
        coordinate=[19, 0.5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B21 = Bus(
        name="B21",
        coordinate=[20, 0.5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B22 = Bus(
        name="B22",
        coordinate=[21, 0.5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B23 = Bus(
        name="B23",
        coordinate=[22, 0.5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B24 = Bus(
        name="B24",
        coordinate=[23, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B25 = Bus(
        name="B25",
        coordinate=[24, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B26 = Bus(
        name="B26",
        coordinate=[25, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B27 = Bus(
        name="B27",
        coordinate=[26, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B28 = Bus(
        name="B28",
        coordinate=[3, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B29 = Bus(
        name="B29",
        coordinate=[4, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B30 = Bus(
        name="B30",
        coordinate=[5, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B31 = Bus(
        name="B31",
        coordinate=[6, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B32 = Bus(
        name="B32",
        coordinate=[7, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B33 = Bus(
        name="B33",
        coordinate=[8, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B34 = Bus(
        name="B34",
        coordinate=[9, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B35 = Bus(
        name="B35",
        coordinate=[10, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B36 = Bus(
        name="B36",
        coordinate=[3, 2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B37 = Bus(
        name="B37",
        coordinate=[4, 2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B38 = Bus(
        name="B38",
        coordinate=[5, 2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B39 = Bus(
        name="B39",
        coordinate=[6, 2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B40 = Bus(
        name="B40",
        coordinate=[7, 2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B41 = Bus(
        name="B41",
        coordinate=[8, 2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B42 = Bus(
        name="B42",
        coordinate=[9, 2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B43 = Bus(
        name="B43",
        coordinate=[10, 2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B44 = Bus(
        name="B44",
        coordinate=[11, 2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B45 = Bus(
        name="B45",
        coordinate=[12, 2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B46 = Bus(
        name="B46",
        coordinate=[13, 2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B47 = Bus(
        name="B47",
        coordinate=[4, -1.5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B48 = Bus(
        name="B48",
        coordinate=[5, -1.5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B49 = Bus(
        name="B49",
        coordinate=[6, -1.5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B50 = Bus(
        name="B50",
        coordinate=[7, -1.5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B51 = Bus(
        name="B51",
        coordinate=[8, 1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B52 = Bus(
        name="B52",
        coordinate=[9, 1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B53 = Bus(
        name="B53",
        coordinate=[9, -1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B54 = Bus(
        name="B54",
        coordinate=[10, -1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B55 = Bus(
        name="B55",
        coordinate=[11, -1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B56 = Bus(
        name="B56",
        coordinate=[12, -1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B57 = Bus(
        name="B57",
        coordinate=[13, -1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B58 = Bus(
        name="B58",
        coordinate=[14, -1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B59 = Bus(
        name="B59",
        coordinate=[15, -1.5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B60 = Bus(
        name="B60",
        coordinate=[16, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B61 = Bus(
        name="B61",
        coordinate=[17, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B62 = Bus(
        name="B62",
        coordinate=[18, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B63 = Bus(
        name="B63",
        coordinate=[19, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B64 = Bus(
        name="B64",
        coordinate=[20, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B65 = Bus(
        name="B65",
        coordinate=[21, -2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B66 = Bus(
        name="B66",
        coordinate=[11, 1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B67 = Bus(
        name="B67",
        coordinate=[12, 1],
        fail_rate_per_year=fail_rate_trafo,
    )

    # Microgrid:
    B68 = Bus(
        name="B68",
        coordinate=[16, 1.5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B69 = Bus(
        name="B69",
        coordinate=[17, 1.5],
        fail_rate_per_year=fail_rate_trafo,
    )
    B70 = Bus(
        name="B70",
        coordinate=[17, 2],
        fail_rate_per_year=fail_rate_trafo,
    )
    B71 = Bus(
        name="B71",
        coordinate=[17, 1],
        fail_rate_per_year=fail_rate_trafo,
    )

    Battery(name="Bat1", bus=B68)
    Production(name="P1", bus=B70)
    Production(name="P2", bus=B71)

    # Lines, connections and impedances
    L1 = Line(
        name="L1",
        fbus=B1,
        tbus=B2,
        r=0.0005,
        x=0.0012,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L2 = Line(
        name="L2",
        fbus=B2,
        tbus=B3,
        r=0.0005,
        x=0.0012,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L3 = Line(
        name="L3",
        fbus=B3,
        tbus=B4,
        r=0.0015,
        x=0.0036,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L4 = Line(
        name="L4",
        fbus=B4,
        tbus=B5,
        r=0.0251,
        x=0.0294,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L5 = Line(
        name="L5",
        fbus=B5,
        tbus=B6,
        r=0.3660,
        x=0.1864,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L6 = Line(
        name="L6",
        fbus=B6,
        tbus=B7,
        r=0.3811,
        x=0.1941,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L7 = Line(
        name="L7",
        fbus=B7,
        tbus=B8,
        r=0.0922,
        x=0.0470,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L8 = Line(
        name="L8",
        fbus=B8,
        tbus=B9,
        r=0.0493,
        x=0.0251,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L9 = Line(
        name="L9",
        fbus=B9,
        tbus=B10,
        r=0.8190,
        x=0.2707,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L10 = Line(
        name="L10",
        fbus=B10,
        tbus=B11,
        r=0.1872,
        x=0.0619,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L11 = Line(
        name="L11",
        fbus=B11,
        tbus=B12,
        r=0.7114,
        x=0.2351,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L12 = Line(
        name="L12",
        fbus=B12,
        tbus=B13,
        r=1.0300,
        x=0.3400,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L13 = Line(
        name="L13",
        fbus=B13,
        tbus=B14,
        r=1.0440,
        x=0.3450,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L14 = Line(
        name="L14",
        fbus=B14,
        tbus=B15,
        r=1.0580,
        x=0.3496,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L15 = Line(
        name="L15",
        fbus=B15,
        tbus=B16,
        r=0.1966,
        x=0.0650,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L16 = Line(
        name="L16",
        fbus=B16,
        tbus=B17,
        r=0.3744,
        x=0.1238,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L17 = Line(
        name="L17",
        fbus=B17,
        tbus=B18,
        r=0.0047,
        x=0.0016,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L18 = Line(
        name="L18",
        fbus=B18,
        tbus=B19,
        r=0.3276,
        x=0.1083,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L19 = Line(
        name="L19",
        fbus=B19,
        tbus=B20,
        r=0.2106,
        x=0.0690,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L20 = Line(
        name="L20",
        fbus=B20,
        tbus=B21,
        r=0.3416,
        x=0.1129,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L21 = Line(
        name="L21",
        fbus=B21,
        tbus=B22,
        r=0.0140,
        x=0.0046,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L22 = Line(
        name="L22",
        fbus=B22,
        tbus=B23,
        r=0.1591,
        x=0.0526,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L23 = Line(
        name="L23",
        fbus=B23,
        tbus=B24,
        r=0.3463,
        x=0.1145,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L24 = Line(
        name="L24",
        fbus=B24,
        tbus=B25,
        r=0.7488,
        x=0.2475,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L25 = Line(
        name="L25",
        fbus=B25,
        tbus=B26,
        r=0.3089,
        x=0.1021,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L26 = Line(
        name="L26",
        fbus=B26,
        tbus=B27,
        r=0.1732,
        x=0.0572,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L27 = Line(
        name="L27",
        fbus=B3,
        tbus=B28,
        r=0.0044,
        x=0.0108,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L28 = Line(
        name="L28",
        fbus=B28,
        tbus=B29,
        r=0.0640,
        x=0.1565,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L29 = Line(
        name="L29",
        fbus=B29,
        tbus=B30,
        r=0.3978,
        x=0.1315,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L30 = Line(
        name="L30",
        fbus=B30,
        tbus=B31,
        r=0.0702,
        x=0.0232,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L31 = Line(
        name="L31",
        fbus=B31,
        tbus=B32,
        r=0.3510,
        x=0.1160,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L32 = Line(
        name="L32",
        fbus=B32,
        tbus=B33,
        r=0.8390,
        x=0.2816,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L33 = Line(
        name="L33",
        fbus=B33,
        tbus=B34,
        r=1.7080,
        x=0.5646,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L34 = Line(
        name="L34",
        fbus=B34,
        tbus=B35,
        r=1.4740,
        x=0.4873,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L35 = Line(
        name="L35",
        fbus=B3,
        tbus=B36,
        r=0.0044,
        x=0.0108,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L36 = Line(
        name="L36",
        fbus=B36,
        tbus=B37,
        r=0.0640,
        x=0.11565,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L37 = Line(
        name="L37",
        fbus=B37,
        tbus=B38,
        r=0.1053,
        x=0.1230,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L38 = Line(
        name="L38",
        fbus=B38,
        tbus=B39,
        r=0.0304,
        x=0.0355,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L39 = Line(
        name="L39",
        fbus=B39,
        tbus=B40,
        r=0.0018,
        x=0.0021,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L40 = Line(
        name="L40",
        fbus=B40,
        tbus=B41,
        r=0.7283,
        x=0.8509,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L41 = Line(
        name="L41",
        fbus=B41,
        tbus=B42,
        r=0.31,
        x=0.3623,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L42 = Line(
        name="L42",
        fbus=B42,
        tbus=B43,
        r=0.0410,
        x=0.0478,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L43 = Line(
        name="L43",
        fbus=B43,
        tbus=B44,
        r=0.0092,
        x=0.0116,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L44 = Line(
        name="L44",
        fbus=B44,
        tbus=B45,
        r=0.1089,
        x=0.1373,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L45 = Line(
        name="L45",
        fbus=B45,
        tbus=B46,
        r=0.0009,
        x=0.0012,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L46 = Line(
        name="L46",
        fbus=B4,
        tbus=B47,
        r=0.0034,
        x=0.0084,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L47 = Line(
        name="L47",
        fbus=B47,
        tbus=B48,
        r=0.0851,
        x=0.2083,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L48 = Line(
        name="L48",
        fbus=B48,
        tbus=B49,
        r=0.2898,
        x=0.7091,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L49 = Line(
        name="L49",
        fbus=B49,
        tbus=B50,
        r=0.0822,
        x=0.2011,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L50 = Line(
        name="L50",
        fbus=B8,
        tbus=B51,
        r=0.0928,
        x=0.0473,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L51 = Line(
        name="L51",
        fbus=B51,
        tbus=B52,
        r=0.3319,
        x=0.1114,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L52 = Line(
        name="L52",
        fbus=B9,
        tbus=B53,
        r=0.1740,
        x=0.0886,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L53 = Line(
        name="L53",
        fbus=B53,
        tbus=B54,
        r=0.2030,
        x=0.1034,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L54 = Line(
        name="L54",
        fbus=B54,
        tbus=B55,
        r=0.2842,
        x=0.1447,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L55 = Line(
        name="L55",
        fbus=B55,
        tbus=B56,
        r=0.2813,
        x=0.1433,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L56 = Line(
        name="L56",
        fbus=B56,
        tbus=B57,
        r=1.5900,
        x=0.5337,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L57 = Line(
        name="L57",
        fbus=B57,
        tbus=B58,
        r=0.7837,
        x=0.2630,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L58 = Line(
        name="L58",
        fbus=B58,
        tbus=B59,
        r=0.3042,
        x=0.1006,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L59 = Line(
        name="L59",
        fbus=B59,
        tbus=B60,
        r=0.3861,
        x=0.1172,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L60 = Line(
        name="L60",
        fbus=B60,
        tbus=B61,
        r=0.5075,
        x=0.2585,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L61 = Line(
        name="L61",
        fbus=B61,
        tbus=B62,
        r=0.0974,
        x=0.0496,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L62 = Line(
        name="L62",
        fbus=B62,
        tbus=B63,
        r=0.1450,
        x=0.0738,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L63 = Line(
        name="L63",
        fbus=B63,
        tbus=B64,
        r=0.7105,
        x=0.3619,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L64 = Line(
        name="L64",
        fbus=B64,
        tbus=B65,
        r=1.0410,
        x=0.5302,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L65 = Line(
        name="L65",
        fbus=B11,
        tbus=B66,
        r=0.2012,
        x=0.0611,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L66 = Line(
        name="L66",
        fbus=B66,
        tbus=B67,
        r=0.0047,
        x=0.0014,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )

    # Microgrid:
    L67 = Line(
        name="L67",
        fbus=B16,
        tbus=B68,
        r=0.7394,
        x=0.2444,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    L68 = Line(
        name="L68",
        fbus=B68,
        tbus=B69,
        r=0.0047,
        x=0.0016,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    ML3 = Line(
        name="ML3",
        fbus=B68,
        tbus=B70,
        r=0.0047,
        x=0.0016,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )
    ML4 = Line(
        name="ML4",
        fbus=B68,
        tbus=B71,
        r=0.0047,
        x=0.0016,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
    )

    # Backup lines:
    L69 = Line(
        name="L69",
        fbus=B11,
        tbus=B43,
        r=0.3042,
        x=0.1006,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
        capacity=5,
    )
    L70 = Line(
        name="L70",
        fbus=B17,
        tbus=B24,
        r=0.8390,
        x=0.2816,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
        capacity=5,
    )
    # L71 = Line(
    #     "L71",
    #     B15,
    #     B46,
    #     r=0.0,
    #     x=0.0,
    #     fail_rate_density_per_year=fail_rate_line,
    #     capacity=3,
    # )
    L72 = Line(
        name="L72",
        fbus=B50,
        tbus=B59,
        r=0.2012,
        x=0.0611,
        fail_rate_density_per_year=fail_rate_line,
        repair_time_dist=line_stat_dist,
        capacity=5,
    )
    # L73 = Line(
    #     "L73",
    #     B27,
    #     B65,
    #     r=0.0,
    #     x=0.0,
    #     fail_rate_density_per_year=fail_rate_line,
    #     capacity=3,
    # )

    CircuitBreaker("E1", L1)
    CircuitBreaker("E2", L67)

    Disconnector("L1a", L1, B1)
    Disconnector("L1b", L1, B2)
    Disconnector("L2a", L2, B2)
    Disconnector("L2b", L2, B3)
    Disconnector("L3a", L3, B3)
    Disconnector("L3b", L3, B4)
    Disconnector("L4a", L4, B4)
    Disconnector("L4b", L4, B5)
    Disconnector("L5a", L5, B5)
    Disconnector("L5b", L5, B6)
    Disconnector("L6a", L6, B6)
    Disconnector("L6b", L6, B7)
    Disconnector("L7a", L7, B7)
    Disconnector("L7b", L7, B8)
    Disconnector("L8a", L8, B8)
    Disconnector("L8b", L8, B9)
    Disconnector("L9a", L9, B9)
    Disconnector("L9b", L9, B10)
    Disconnector("L10a", L10, B10)
    Disconnector("L10b", L10, B11)
    Disconnector("L11a", L11, B11)
    Disconnector("L11b", L11, B12)
    Disconnector("L12a", L12, B12)
    Disconnector("L12b", L12, B13)
    Disconnector("L13a", L13, B13)
    Disconnector("L13b", L13, B14)
    Disconnector("L14a", L14, B14)
    Disconnector("L14b", L14, B15)
    Disconnector("L15a", L15, B15)
    Disconnector("L15b", L15, B16)
    Disconnector("L16a", L16, B16)
    Disconnector("L16b", L16, B17)
    Disconnector("L17a", L17, B17)
    Disconnector("L17b", L17, B18)
    Disconnector("L18a", L18, B18)
    Disconnector("L18b", L18, B19)
    Disconnector("L19a", L19, B19)
    Disconnector("L19b", L19, B20)
    Disconnector("L20a", L20, B20)
    Disconnector("L20b", L20, B21)
    Disconnector("L21a", L21, B21)
    Disconnector("L21b", L21, B22)
    Disconnector("L22a", L22, B22)
    Disconnector("L22b", L22, B23)
    Disconnector("L23a", L23, B23)
    Disconnector("L23b", L23, B24)
    Disconnector("L24a", L24, B24)
    Disconnector("L24b", L24, B25)
    Disconnector("L25a", L25, B25)
    Disconnector("L25b", L25, B26)
    Disconnector("L26a", L26, B26)
    Disconnector("L26b", L26, B27)

    Disconnector("L27a", L27, B3)
    Disconnector("L27b", L27, B28)
    Disconnector("L28a", L28, B28)
    Disconnector("L28b", L28, B29)
    Disconnector("L29a", L29, B29)
    Disconnector("L29b", L29, B30)
    Disconnector("L30a", L30, B30)
    Disconnector("L30b", L30, B31)
    Disconnector("L31a", L31, B31)
    Disconnector("L31b", L31, B32)
    Disconnector("L32a", L32, B32)
    Disconnector("L32b", L32, B33)
    Disconnector("L33a", L33, B33)
    Disconnector("L33b", L33, B34)
    Disconnector("L34a", L34, B34)
    Disconnector("L34b", L34, B35)

    Disconnector("L35a", L35, B3)
    Disconnector("L35b", L35, B36)
    Disconnector("L36a", L36, B36)
    Disconnector("L36b", L36, B37)
    Disconnector("L37a", L37, B37)
    Disconnector("L37b", L37, B38)
    Disconnector("L38a", L38, B38)
    Disconnector("L38b", L38, B39)
    Disconnector("L39a", L39, B39)
    Disconnector("L39b", L39, B40)
    Disconnector("L40a", L40, B40)
    Disconnector("L40b", L40, B41)
    Disconnector("L41a", L41, B41)
    Disconnector("L41b", L41, B42)
    Disconnector("L42a", L42, B42)
    Disconnector("L42b", L42, B43)
    Disconnector("L43a", L43, B43)
    Disconnector("L43b", L43, B44)
    Disconnector("L44a", L44, B44)
    Disconnector("L44b", L44, B45)
    Disconnector("L45a", L45, B45)
    Disconnector("L45b", L45, B46)

    Disconnector("L46a", L46, B4)
    Disconnector("L46b", L46, B47)
    Disconnector("L47a", L47, B47)
    Disconnector("L47b", L47, B48)
    Disconnector("L48a", L48, B48)
    Disconnector("L48b", L48, B49)
    Disconnector("L49a", L49, B49)
    Disconnector("L49b", L49, B50)

    Disconnector("L50a", L50, B8)
    Disconnector("L50b", L50, B51)
    Disconnector("L51a", L51, B51)
    Disconnector("L51b", L51, B52)

    Disconnector("L52a", L52, B9)
    Disconnector("L52b", L52, B53)
    Disconnector("L53a", L53, B53)
    Disconnector("L53b", L53, B54)
    Disconnector("L54a", L54, B54)
    Disconnector("L54b", L54, B55)
    Disconnector("L55a", L55, B55)
    Disconnector("L55b", L55, B56)
    Disconnector("L56a", L56, B56)
    Disconnector("L56b", L56, B57)
    Disconnector("L57a", L57, B57)
    Disconnector("L57b", L57, B58)

    Disconnector("L58a", L58, B58)
    Disconnector("L58b", L58, B59)
    Disconnector("L59a", L59, B59)
    Disconnector("L59b", L59, B60)
    Disconnector("L60a", L60, B60)
    Disconnector("L60b", L60, B61)
    Disconnector("L61a", L61, B61)
    Disconnector("L61b", L61, B62)
    Disconnector("L62a", L62, B62)
    Disconnector("L62b", L62, B63)
    Disconnector("L63a", L63, B63)
    Disconnector("L63b", L63, B64)
    Disconnector("L64a", L64, B64)
    Disconnector("L64b", L64, B65)

    Disconnector("L65a", L65, B11)
    Disconnector("L65b", L65, B66)
    Disconnector("L66a", L66, B66)
    Disconnector("L66b", L66, B67)

    # Microgrid:
    Disconnector("L67a", L67, B16)
    Disconnector("L67b", L67, B68)
    Disconnector("L68a", L68, B68)
    Disconnector("L68b", L68, B69)
    Disconnector("ML3a", ML3, B68)
    Disconnector("ML3b", ML3, B70)
    Disconnector("ML4a", ML4, B68)
    Disconnector("ML4b", ML4, B71)

    # Backup lines:
    Disconnector("L69a", L69, B11)
    Disconnector("L69b", L69, B43)
    Disconnector("L70a", L70, B17)
    Disconnector("L70b", L70, B24)
    # Disconnector("L71a", L71, B15)
    # Disconnector("L71b", L71, B46)
    Disconnector("L72a", L72, B50)
    Disconnector("L72b", L72, B59)
    # Disconnector("L73a", L73, B27)
    # Disconnector("L73b", L73, B65)

    L69.set_backup()
    L70.set_backup()
    # L71.set_backup()
    L72.set_backup()
    # L73.set_backup()

    tn = Transmission(ps, B1)

    dn = Distribution(tn, L1)

    dn.add_buses(
        [
            B2,
            B3,
            B4,
            B5,
            B6,
            B7,
            B8,
            B9,
            B10,
            B11,
            B12,
            B13,
            B14,
            B15,
            B16,
            B17,
            B18,
            B19,
            B20,
            B21,
            B22,
            B23,
            B24,
            B25,
            B26,
            B27,
            B28,
            B29,
            B30,
            B31,
            B32,
            B33,
            B34,
            B35,
            B36,
            B37,
            B38,
            B39,
            B40,
            B41,
            B42,
            B43,
            B44,
            B45,
            B46,
            B47,
            B48,
            B49,
            B50,
            B51,
            B52,
            B53,
            B54,
            B55,
            B56,
            B57,
            B58,
            B59,
            B60,
            B61,
            B62,
            B63,
            B64,
            B65,
            B66,
            B67,
            # B68,
            # B69,
        ]
    )

    dn.add_lines(
        [
            L2,
            L3,
            L4,
            L5,
            L6,
            L7,
            L8,
            L9,
            L10,
            L11,
            L12,
            L13,
            L14,
            L15,
            L16,
            L17,
            L18,
            L19,
            L20,
            L21,
            L22,
            L23,
            L24,
            L25,
            L26,
            L27,
            L28,
            L29,
            L30,
            L31,
            L32,
            L33,
            L34,
            L35,
            L36,
            L37,
            L38,
            L39,
            L40,
            L41,
            L42,
            L43,
            L44,
            L45,
            L46,
            L47,
            L48,
            L49,
            L50,
            L51,
            L52,
            L53,
            L54,
            L55,
            L56,
            L57,
            L58,
            L59,
            L60,
            L61,
            L62,
            L63,
            L64,
            L65,
            L66,
            # L67,
            # L68,
            L69,
            L70,
            # L71,
            L72,
            # L73,
        ]
    )

    m = Microgrid(dn, L67, mode=microgrid_mode)
    m.add_buses([B68, B69, B70, B71])
    m.add_lines([L68, ML3, ML4])

    return ps


if __name__ == "__main__":
    import os

    ps = initialize_network()
    fig = plot_topology(ps.buses, ps.lines)

    fig.savefig(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "IEEE69.pdf")
    )



from relsad.network.components import Bus, Line

from .PowerNetwork import PowerNetwork


class Transmission(PowerNetwork):
    """Class defining a transmission network type

    ...

    Attributes
    ----------
    name : str
        Name of the transmission system
    parent_network : PowerSystem
        The parent network of the transmission network
    child_network_list : list
        Lists of child networks
    trafo_bus : Bus
        Bus with tranformer connected to distribution network
    buses : list
        List with the buses in the transmission network
    sections : list
        List with the sections in the transmission network
    ev_parks : list
        List with the EV parks in the transmission network
    p_energy_shed : float
        The active power energy.shed in the transmission network
    acc_p_energy_shed : float
        The accumulated active power energy.shedding in the transmission network
    q_energy_shed : float
        The reactive power energy.shed in the transmission network
    acc_q_energy_shed : float
        The accumulated reactive power energy.shedding in the transmission network
    history : dict
        Dictionary containing the history variables of the transmission network
    monte_carlo_history : dict
        Dictionary containing the history variables from the monte carlo simulation



    Methods
    ----------
    get_trafo_bus()
        Returns the bus connecting to the overlying network (transmission network)
    reset_slack_bus()
        Resets the slack bus of the transmission network
    add_chil_network(network)
        Adds child network
    get_lines()
        Returns the lines in the transmission network
    get_monte_carlo_history(attribute)
        Returns the specified history variable from the Monte Carlo simulation
    get_history(attribute)
        Returns the specified history variable
    get_system_load()
        Returns the system load at the current time in MW and MVar
    reset_energy_shed_variables()
        Resets the energy.shed variables

    """

    ## Visual attributes
    color = "steelblue"

    ## Counter
    counter = 0

    def __init__(self, parent_network, trafo_bus: Bus):
        Transmission.counter += 1
        self.name = "trans_network{:d}".format(Transmission.counter)

        self.parent_network = parent_network
        parent_network.add_child_network(self)
        self.child_network_list = list()

        self.trafo_bus = trafo_bus
        self.buses = [trafo_bus]
        self.sections = list()

        self.ev_parks = list()

        trafo_bus.handle.color = self.color
        trafo_bus.color = self.color
        self.parent_network.add_bus(trafo_bus)

        trafo_bus.set_slack()
        parent_network.slack = trafo_bus
        # Load shedding
        self.p_energy_shed = 0
        self.acc_p_energy_shed = 0
        self.q_energy_shed = 0
        self.acc_q_energy_shed = 0

        ## History
        self.history: dict = {}
        self.monte_carlo_history: dict = {}

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Transmission(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def get_trafo_bus(self):
        """
        Returns the bus connecting to the overlying network (transmission network)

        Parameters
        ----------
        None

        Returns
        ----------
        bus : Bus
            The bus connecting the overlying network

        """
        return self.trafo_bus

    def reset_slack_bus(self):
        """
        Resets the slack bus of the transmission network

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.trafo_bus.set_slack()

    def add_child_network(self, network):
        """
        Adds child network

        Parameters
        ----------
        network : PowerNetwork
            A network

        Returns
        ----------
        None

        """
        self.child_network_list.append(network)
        self.parent_network.add_child_network(network)

    def add_bus(self, bus: Bus):
        pass

    def add_buses(self, buses: list):
        pass

    def add_line(self, line: Line):
        pass

    def add_lines(self, lines: list):
        pass

    def get_lines(self):
        """
        Returns the lines in the transmission network

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        return None

    def get_monte_carlo_history(self, attribute):
        """
        Returns the specified history variable

        Parameters
        ----------
        attribute : str
            System attribute

        Returns
        ----------
        monte_carlo_history[attribute] : dict
            Returns the history variables of an attribute from the Monte Carlo simulation

        """
        return self.monte_carlo_history[attribute]

    def get_history(self, attribute):
        """

        Returns the specified history variable

        Parameters
        ----------
        attribute : str
            System attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute

        """
        return self.history[attribute]

    def get_system_load(self):
        """
        Returns the system load at the current time in MW and MVar

        Parameters
        ----------
        None

        Returns
        ----------
        pload : float
            The active power load in the transmission system
        qload : float
            The reactive power load in the tranmission system

        """
        pload, qload = 0, 0
        for bus in self.buses:
            p, q = bus.get_load()
            pload += p
            qload += q
        return pload, qload

    def reset_energy_shed_variables(self):
        """
        Resets the energy.shed variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.p_energy_shed = 0
        self.acc_p_energy_shed = 0
        self.q_energy_shed = 0
        self.acc_q_energy_shed = 0



import itertools

from relsad.network.components import Line
from relsad.network.containers import Section
from relsad.utils import unique


def get_downstream_lines(parent_line: Line):
    """
    Function that finds and returns all downstream lines in a radial tree

    Parameters
    ----------
    parent_line : Line
        Parent line of radial tree

    Returns
    -------
    downstream_lines : list
        List of all downstream lines from parent_line

    """
    lines = [
        x
        for x in parent_line.tbus.fromline_list
        if x in parent_line.parent_network.lines
    ]
    if len(lines) == 0:
        return [[parent_line]]
    downstream_lines = []
    for nline in lines:
        for downstream_line in get_downstream_lines(nline):
            downstream_lines.append([parent_line] + downstream_line)
    return downstream_lines


def create_downstream_sections(
    curr_line: Line,
    parent_section: Section = None,
):
    """
    Creates sections downstream from the current line

    Parameters
    ----------
    curr_line : Line
        The current line
    parent_section : Section
        The parent section

    Returns
    -------
    parent_section : Section
        The updated parent section

    """
    if parent_section is None:
        # No parent section, gathers all downstream lines and
        # switches to create a parent section
        lines = unique(
            itertools.chain.from_iterable(get_downstream_lines(curr_line))
        )
        parent_section = Section(
            parent_section=None,
            lines=lines,
            switches=curr_line.get_switches(),
        )
        parent_section = create_downstream_sections(
            curr_line=curr_line, parent_section=parent_section
        )
    else:
        # A parent section exists, gathers all downstream lines and
        # switches to create a downstream child section
        next_lines = [
            x
            for x in curr_line.tbus.fromline_list
            if x in curr_line.parent_network.lines
        ]
        if next_lines != []:
            # Not an end line, continues searching for
            # switches in downstream lines
            for next_line in next_lines:
                if (
                    next_line.disconnectors == []
                    and next_line.circuitbreaker is None
                ):
                    # Line without a switch, continue
                    # searching downstream
                    parent_section = create_downstream_sections(
                        curr_line=next_line, parent_section=parent_section
                    )
                else:
                    # Line with a switch, gather downstream
                    # lines and switches for a new child section
                    lines = unique(
                        itertools.chain.from_iterable(
                            get_downstream_lines(next_line)
                        )
                    )
                    section = Section(
                        parent_section=parent_section,
                        lines=lines,
                        switches=next_line.get_switches(),
                    )
                    section = create_downstream_sections(
                        curr_line=next_line, parent_section=section
                    )
                    parent_section.add_child_section(section)
                    if len(next_line.get_switches()) > 1 and lines != [
                        next_line
                    ]:
                        # Multiple switches on line (that is not an end line)
                        # Add a child section containing line and switches
                        child_section = Section(
                            parent_section=section,
                            lines=[next_line],
                            switches=next_line.get_switches(),
                        )
                        section.add_child_section(child_section)
    return parent_section


def refine_sections(
    parent_section: Section,
):
    """
    Refines the sections within the parent section including the parent
    section itself. Removes unnecessary coarse sections. Returns the
    refined parent section.

    Parameters
    ----------
    parent_section : Section
        The parent section

    Returns
    -------
    parent_section : Section
        The refined parent section

    """
    # Finds the unique lines of the parent section
    # not shared by any of the child sections
    child_lines = set(
        itertools.chain.from_iterable(
            [x.lines for x in parent_section.child_sections]
        )
    )
    lines = list(set(parent_section.lines) - child_lines)

    # Creates internal sections recursively in
    # the child sections
    child_sections = parent_section.child_sections
    for child_section in child_sections:
        child_section = refine_sections(child_section)

    if lines == [] and parent_section.parent_section is not None:
        # No unique parent section lines, parent section
        # covered by its child sections and thus unnecessary
        # Removes parent section
        parent_section.parent_section.child_sections.remove(parent_section)
        # Adds current child sections to the parent of the removed
        # parent section
        parent_section = parent_section.parent_section
        for child_section in child_sections:
            if child_section not in parent_section.child_sections:
                parent_section.child_sections.append(child_section)
        parent_section = refine_sections(parent_section)
    else:
        # Parent section has unique lines
        # Remove non-unique lines and switches from
        # parent section
        parent_section.lines = lines
        parent_section.switches = unique(
            parent_section.switches
            + [
                x
                for child_section in parent_section.child_sections
                for x in child_section.switches
                if x.line in parent_section.lines
                or sum(
                    [
                        toline in parent_section.lines
                        for toline in x.line.fbus.toline_list
                    ]
                )
                > 0
            ]
        )
    if len(parent_section.lines) == 1:
        if len(parent_section.lines[0].get_switches()) > 0:
            # Single line parent section with switches
            parent_section.switches = parent_section.lines[0].get_switches()
    return parent_section


def get_section_list(
    parent_section: Section,
    section_list: list = [],
):
    """
    Appends and returns a list containing the sections in the path

    Parameters
    ----------
    parent_section : Section
        The parent section
    section_list : list
        List containing the sections

    Returns
    -------
    section_list : list
        List containing the sections

    """
    if section_list == []:
        section_list.append(parent_section)
    section_list += parent_section.child_sections
    for child in parent_section.child_sections:
        section_list = get_section_list(
            parent_section=child,
            section_list=section_list,
        )
    return section_list


def create_sections(connected_line: Line):
    """
    Create layered network sections starting downstream from the network connected line.
    The sections are separated by disconnectors.

    Parameters
    ----------
    connected_line : Line
        The line connecting the network to the parent network

    Returns
    -------
    parent_section : Section
        The parent section of the network

    """
    parent_section = create_downstream_sections(
        curr_line=connected_line,
        parent_section=None,
    )
    parent_section = refine_sections(parent_section=parent_section)
    parent_section.attach_to_lines()
    return parent_section



import copy
import os

from relsad.network.systems import PowerSystem
from relsad.results.storage import save_monte_carlo_history_from_dict


def save_network_monte_carlo_history(
    power_system: PowerSystem,
    save_dir: str,
    save_dict: dict,
):
    """
    Saves the history of the energy.shedding in the power system

    Parameters
    ----------
    power_system : PowerSystem
        A power system element
    save_dir : str
        The saving path
    save_dict : dict
        Dictionary with simulation results

    Returns
    ----------
    None

    """
    if not os.path.isdir(save_dir):
        os.makedirs(save_dir)
    network_state_list = list(save_dict[power_system.name].keys())
    power_system_save_dir = os.path.join(save_dir, power_system.name)
    for state_var in network_state_list:
        save_monte_carlo_history_from_dict(
            save_dict, [power_system], state_var, power_system_save_dir
        )
    for network in power_system.child_network_list:
        network_save_dir = os.path.join(save_dir, network.name)
        for state_var in network_state_list:
            save_monte_carlo_history_from_dict(
                save_dict, [network], state_var, network_save_dir
            )
    for bus in power_system.buses:
        bus_save_dir = os.path.join(save_dir, bus.name)
        bus_state_list = list(save_dict[bus.name].keys())
        for state_var in bus_state_list:
            save_monte_carlo_history_from_dict(
                save_dict, [bus], state_var, bus_save_dir
            )
        if bus.ev_park is not None:
            ev_park_save_dir = os.path.join(bus_save_dir, bus.ev_park.name)
            ev_park_state_list = list(save_dict[bus.ev_park.name].keys())
            for state_var in ev_park_state_list:
                save_monte_carlo_history_from_dict(
                    save_dict, [bus.ev_park], state_var, ev_park_save_dir
                )


def merge_monte_carlo_history(
    power_system: PowerSystem,
    iteration_dicts: list,
):
    """
    Merges the Monte Carlo history from all the iterations in the simulation

    Parameters
    ----------
    power_system : PowerSystem
        A power system element
    iteration_dicts : list
        List containing information about all the iterations

    Returns
    ----------
    save_dict : dict
        Dictionary with simulation results

    """
    save_dict = copy.deepcopy(iteration_dicts[0])
    for it_dict in iteration_dicts:
        curr_it_dict = it_dict[power_system.name]
        network_state_list = list(curr_it_dict.keys())
        it = list(curr_it_dict[network_state_list[0]].keys())[0]
        for state_var in network_state_list:
            save_dict[power_system.name][state_var][it] = (
                curr_it_dict[state_var][it]
            )
        save_dict = merge_monte_carlo_child_network_history(
            power_system, it_dict, it, save_dict
        )
        save_dict = merge_monte_carlo_comp_history(
            power_system, it_dict, it, save_dict
        )
    return save_dict


def merge_monte_carlo_child_network_history(
    power_system: PowerSystem,
    it_dict: dict,
    it: int,
    save_dict: dict,
):
    """
    Merges the lists used for history variables from the Monte Carlo simulation

    Parameters
    ----------
    power_system : PowerSystem
        A power system element
    it_dict : dict
        Dictionary with iteration results
    it : int
        Iteration number
    save_dict : dict
        Dictionary with simulation results

    Returns
    ----------
    save_dict : dict
        Dictionary with simulation results

    """
    for network in power_system.child_network_list:
        curr_it_dict = it_dict[network.name]
        network_state_list = list(curr_it_dict.keys())
        for state_var in network_state_list:
            save_dict[network.name][state_var][it] = (
                curr_it_dict[state_var][it]
            )
    return save_dict


def merge_monte_carlo_comp_history(
    power_system: PowerSystem,
    it_dict: dict,
    it: int,
    save_dict: dict,
):
    """
    Merges the lists used for component history variables
    from the Monte Carlo simulation

    Parameters
    ----------
    power_system : PowerSystem
        A power system element
    it_dict : dict
        Dictionary with iteration results
    it : int
        The iteration number
    save_dict : dict
        Dictionary with simulation results

    Returns
    ----------
    save_dict : dict
        Dictionary with simulation results

    """
    for bus in power_system.buses:
        curr_bus_it_dict = it_dict[bus.name]
        bus_state_list = list(curr_bus_it_dict.keys())
        for state_var in bus_state_list:
            save_dict[bus.name][state_var][it] = (
                curr_bus_it_dict[state_var][it]
            )
        if bus.ev_park is not None:
            curr_ev_park_it_dict = it_dict[bus.ev_park.name]
            ev_park_state_list = list(curr_ev_park_it_dict.keys())
            for state_var in ev_park_state_list:
                save_dict[bus.ev_park.name][state_var][it] = (
                    curr_ev_park_it_dict[state_var][it]
                )
    return save_dict



import matplotlib.lines as mlines
import numpy as np

from relsad.network.containers import SectionState
from relsad.Time import Time, TimeUnit
from relsad.utils import unique

from .Component import Component
from .Line import Line
from .Switch import Switch


class CircuitBreaker(Component, Switch):

    """
    Common base class for circuit breakers

    ...

    Attributes
    ----------
    name : string
        Name of the circuit breaker
    line : Line
        The line the circuit breaker is connected to
    is_open : bool
        Indicates if the circuit breaker is open or closed
    outage_time : Time
        The outage time of the circuit breaker
    coordinate : list
        Coordinate of the circuit breaker
    initial_state : bool
        The initial state of the circuit breaker
    failed : bool
        True if the circuit breaker is in a failed state, False if not
    history : dict
        Dictonary attribute that stores the historic variables

    Methods
    ----------
    close()
        Closes the circuit breaker and the disconnectors connected to the host line
    open()
        Opens the circuit breaker and the disconnectors connected to the host line
    fail()
        Sets the circuit breaker to failed and opens the circuit breaker
    not_fail()
        Sets the circuit breaker to not failed and closes the circuit breaker
    update_fail_status(dt)
        Updates the failure status of the circuit breaker
    initialize_history()
        Initializes the history variables
    update_history(prev_time, curr_time, save_flag)
        Updates the history variables
    get_history(attribute)
        Returns the history variables of an attribute
    add_random_instance(random_gen)
        Adds global random seed
    print_status()
        Prints the status
    reset_status(save_flag)
        Resets and sets the status of the system parameters

    """

    ## Visual attributes
    color = "black"
    edgecolor = "black"
    marker = "s"
    size = 3**2
    handle = mlines.Line2D(
        [],
        [],
        marker=marker,
        markeredgewidth=3,
        markersize=size,
        linestyle="None",
        color=color,
        markeredgecolor=edgecolor,
    )

    ## Random instance
    ps_random: np.random.Generator = None

    def __init__(
        self,
        name: str,
        line: Line,
        is_open: bool = False,
        fail_rate: float = 0.0,
        outage_time: Time = Time(1, TimeUnit.HOUR),
    ):

        # Verify input
        if line is None:
            raise Exception("CircuitBreaker must be connected to a Line")
        if line.parent_network is not None:
            raise Exception(
                "CircuitBreaker must be created before the line is connected to a network"
            )
        if fail_rate < 0:
            raise Exception("The failure rate must be positive")

        self.name = name

        dx = line.tbus.coordinate[0] - line.fbus.coordinate[0]
        dy = line.tbus.coordinate[1] - line.fbus.coordinate[1]
        self.coordinate = [
            line.fbus.coordinate[0] + dx / 3,
            line.fbus.coordinate[1] + dy / 3,
        ]
        self.initial_state = is_open
        self.is_open = is_open
        self.failed = False
        self.fail_rate = fail_rate
        self.outage_time = outage_time
        self.line = line

        # Add this circuitbreaker to parent line
        self.line.circuitbreaker = self

        ## History
        self.history = {}
        self.initialize_history()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"CircuitBreaker(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name and isinstance(
                other, CircuitBreaker
            )
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def close(self):
        """
        Closes the circuit breaker and the disconnectors connected to the circuit breaker

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.is_open = False
        self.color = "black"
        for discon in self.line.disconnectors:
            if (
                discon.is_open
                and self.line.section.state == SectionState.CONNECTED
            ):
                discon.close()
        if not self.line.connected:
            self.line.connect()

    def open(self):
        """
        Opens the circuit breaker and the disconnectors connected to the circuit breaker

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.is_open = True
        self.color = "white"
        for discon in self.line.disconnectors:
            if not discon.is_open:
                discon.open()
        if self.line.connected:
            self.line.disconnect()

    def fail(self):
        """
        Sets the circuit breaker to failed and opens the circuit breaker

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.failed = True
        self.open()

    def not_fail(self):
        """
        Sets the circuit breaker to not failed and closes the circuit breaker

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.failed = False
        self.close()

    def update_fail_status(self, dt: Time):
        """
        Updates the failuer status of the circuit breaker

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """

    def initialize_history(self):
        """
        Initializes the history variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.history["is_open"] = {}

    def update_history(
        self, prev_time: Time, curr_time: Time, save_flag: bool
    ):
        """
        Updates the history variables

        Parameters
        ----------
        prev_time : Time
            The previous time
        curr_time : Time
            Current time
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        if save_flag:
            time = curr_time.get_unit_quantity(curr_time.unit)
            self.history["is_open"][time] = self.is_open

    def get_history(self, attribute: str):
        """
        Returns the history variables of an attribute

        Parameters
        ----------
        attribute : str
            System attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute

        """
        return self.history[attribute]

    def add_random_instance(self, random_gen):
        """
        Adds global random seed

        Parameters
        ----------
        random_gen : int
            Random number generator

        Returns
        ----------
        None

        """
        self.ps_random = random_gen

    def print_status(self):
        """
        Prints the status
        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        pass

    def reset_status(self, save_flag: bool):
        """
        Resets and sets the status of the class parameters

        Parameters
        ----------
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        self.close()
        if save_flag:
            self.initialize_history()


if __name__ == "__main__":
    pass



class CostFunction:
    """
    Class defining a linear energy shedding
    cost function for a bus on the form

    cost = A + B * active_energy_shedded

    ...

    Attributes
    ----------
    A : float
        Constant cost factor parameter
    B : float
        Variable cost factor parameter



    Methods
    ----------
    """

    def __init__(
        self,
        A: float,
        B: float,
    ):
        self.A = A
        self.B = B



from .plot import (
    plot_history,
    plot_monte_carlo_history,
    plot_history_last_state,
)

from .topology import plot_topology

__all__ = []
for v in dir():
    if not v.startswith("__") and v != "relsad":
        __all__.append(v)



from enum import Enum

import matplotlib.lines as mlines
import numpy as np

from relsad.network.containers import SectionState
from relsad.Time import Time, TimeUnit
from relsad.topology.ICT.dfs import is_connected
from relsad.utils import convert_yearly_fail_rate, random_choice, unique

from .Component import Component
from .Controller import Controller, ControllerState
from .ICTNode import ICTNode


class MicrogridMode(Enum):
    """
    Microgrid mode

    Attributes
    ----------
    SURVIVAL : int
        The microgrid is in survival mode, it disconnects
        from the parent power network for the entire failure
        duration
    FULL_SUPPORT : int
        The microgrid is in full support mode, it disconnects
        from the parent power network only during sectioning time.
        When connected to the parent power network, the microgrid
        provides all the support it has available.
    LIMITED_SUPPORT : int
        The microgrid is in limited support mode, it disconnects
        from the parent power network only during sectioning time.
        When connected to the parent power network, the microgrid
        provides only surplus support.
    """

    SURVIVAL = 1
    FULL_SUPPORT = 2
    LIMITED_SUPPORT = 3


class MicrogridController(Component, Controller):
    """
    Common class for microgrid controller

    ...

    Attributes
    ----------
    name : string
        Name of the microgird controller
    ict_node : ICTNode
        The ICT node connected to the controller
    fail_rate : float
        The failure rate of the microgrid controller
    state : ControllerState
        The state of the microgrid controller
    sectioning_time : Time
        The sectioning time in the microgrid
    parent_sectioning_time : Time
        The sectioning time of the parent controller
    check_components : bool
    manual_sectioning_time : Time
        The sectioning time of a passive microgrid without
        controller (or controller is failed)
    power_network : PowerNetwork
        The power network the controller belongs to
    ict_network : ICTNetwork
        The ICT network the controller belongs to
    sensors : list
        List of sensors associated with the microgrid
    failed_sections : list
        List of failed sections in the microgrid
    history : dict
        Dictonary attribute that stores the historic variables
    monte_carlo_history : dict


    Methods
    ----------
    check_circuitbreaker(curr_time, dt)
        Checks if the circuitbreaker in the distribution system is open.
        If sectioning time is finished, disconnect failed sections and close the
        circuitbreaker.
    check_circuitbreaker_manually(curr_time, dt)
        Checks if the circuitbreaker in the distribution system is open manually.
        If sectioning time is finished, disconnect failed sections and close the
        circuitbreaker.
    disconnect_failed_sections()
        Disconnects the failed sections in the microgrid
    check_sensors(curr_time, dt)
        Loops through the sections connected to the controller determining which
        sensors have failed. Performs actions according to the sensor status
        in the respective section.
        If a section was disconnected and no longer includes any failed sensor,
        it is connected.
        If a section was connected and now includes a failed sensor, it is disconnected.
        The total sectioning time is summed from each section.
    run_control_loop(curr_time, dt)
        System control check, determines if components have failed and performes
        the required action
    check_lines_manually(curr_time)
        Loops manually through the sections connected to the controller determining
        which lines have failed. Performs actions according
        to the line status in the respective section.

        If a section was disconnected and no longer includes any failed
        lines, it is connected.

        If a section was connected and now includes a failed line,
        it is disconnected.

        The total sectioning time is summed from each section.
    run_manual_control_loop(curr_time, dt)
        Manual system control check, determines if components have failed and
        performes the required action.
    set_sectioning_time(sectioning_time)
        Sets the sectiom time of the microgrid based on the max value
        of the microgrid sectioning time and the sectioning time set by the
        controller
    set_parent_sectioning_time(sectioning_time)
        Sets the sectioning time of the parent power network
    update_fail_status(dt)
        Updates the fail status
    update_history(prev_time, curr_time, save_flag)
        Updates the history variables
    get_history(attribute)
        Returns the history variables of an attribute
    add_random_instance(random_gen)
        Adds global random seed
    print_status()
        Prints the status
    reset_status(save_flag)
        Resets and sets the status of the class parameters
    initialize_history()
        Initializes the history variables
    """

    def __init__(
        self,
        name: str,
        power_network,
        fail_rate: float = 0,
        state: ControllerState = ControllerState.OK,
    ):

        self.name = name
        self.ict_node = None
        self.fail_rate = fail_rate
        self.state = state
        self.sectioning_time = Time(0)
        self.parent_sectioning_time = Time(0)
        self.check_components = False

        self.manual_sectioning_time = None

        self.power_network = power_network
        self.ict_network = None

        self.sensors = []

        self.failed_sections = []

        ## History
        self.history = {}
        self.monte_carlo_history = {}

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"MicrogridController(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name and isinstance(
                other, MicrogridController
            )
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def check_circuitbreaker(self, curr_time: Time, dt: Time):
        """
        Checks if the circuitbreaker in the distribution system is open.
        If sectioning time is finished, disconnect failed sections and close the
        circuitbreaker.

        Parameters
        ----------
        curr_time : Time
            The current time
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        if self.power_network.connected_line.circuitbreaker.is_open:
            # If circuitbreaker in Microgrid with survival mode and parent Distribution
            # system has no failed lines
            if (
                self.power_network.mode == MicrogridMode.SURVIVAL
                and self.power_network.distribution_network.failed_line is True
            ):
                return
            elif self.sectioning_time <= Time(0):
                self.disconnect_failed_sections()
                if (
                    not self.power_network.connected_line.failed
                    and self.power_network.connected_line
                    not in [
                        line
                        for section in self.failed_sections
                        for line in section.lines
                    ]
                ):
                    # Sectioning time finished
                    self.power_network.connected_line.circuitbreaker.close()
                    self.power_network.connected_line.section.connect(dt, self)
                    self.failed_sections = []

    def check_circuitbreaker_manually(self, curr_time: Time, dt: Time):
        """
        Checks if the circuitbreaker in the distribution system is open manually.
        If sectioning time is finished, disconnect failed sections and close the
        circuitbreaker.

        Parameters
        ----------
        curr_time : Time
            The current time
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        if self.power_network.connected_line.circuitbreaker.is_open:
            # If circuitbreaker in Microgrid with survival mode and parent Distribution
            # system has no failed lines
            if (
                self.power_network.mode == MicrogridMode.SURVIVAL
                and self.power_network.distribution_network.failed_line is True
            ):
                return
            elif self.sectioning_time <= Time(0):
                self.disconnect_failed_sections()
                if not self.power_network.connected_line.failed:
                    # Sectioning time finished
                    self.power_network.connected_line.circuitbreaker.close()
                    self.power_network.connected_line.section.connect_manually()
                    self.failed_sections = []

    def disconnect_failed_sections(self):
        """
        Disconnects the failed sections in the microgrid

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        for section in self.failed_sections:
            section.disconnect()

    def check_sensors(self, curr_time: Time, dt: Time):
        """
        Loops through the sections connected to the controller determining which
        sensors have failed. Performs actions according to the sensor status
        in the respective section.
        If a section was disconnected and no longer includes any failed sensor,
        it is connected.
        If a section was connected and now includes a failed sensor,
        it is disconnected.
        The total sectioning time is summed from each section.

        Parameters
        ----------
        curr_time : Time
            The current time
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        connected_sections = [
            x
            for x in self.power_network.sections
            if x.state == SectionState.CONNECTED
        ]
        disconnected_sections = [
            x
            for x in self.power_network.sections
            if x.state == SectionState.DISCONNECTED
        ]
        for section in disconnected_sections:
            sensors = unique(
                [
                    switch.line.sensor
                    for switch in section.switches
                    if switch.line.sensor is not None
                ]
            )
            num_fails = 0
            need_manual_attention = False
            for sensor in sensors:
                # If no ICT network
                if self.ict_node is None:
                    (
                        repair_time,
                        line_fail_status,
                    ) = sensor.get_line_fail_status(dt)
                    self.sectioning_time += repair_time
                # If both components have ICT nodes
                elif self.ict_node is not None and sensor.ict_node is not None:
                    # If the ICT nodes are connected to each other
                    if is_connected(
                        node_1=self.ict_node,
                        node_2=sensor.ict_node,
                        network=self.ict_network,
                    ):
                        (
                            repair_time,
                            line_fail_status,
                        ) = sensor.get_line_fail_status(dt)
                        self.sectioning_time += repair_time
                    # If the ICT nodes are not connected to each other
                    else:
                        need_manual_attention = True
                        line_fail_status = sensor.line.failed
                # If no ICT node on sensor
                else:
                    need_manual_attention = True
                    line_fail_status = sensor.line.failed
                num_fails += 1 if line_fail_status else 0
            if need_manual_attention is True:
                self.sectioning_time += self.manual_sectioning_time
            if num_fails == 0:
                section.connect(dt, self)
                if section in self.failed_sections:
                    self.failed_sections.remove(section)
        # Loop connected sections
        for section in connected_sections:
            sensors = unique(
                [
                    switch.line.sensor
                    for switch in section.switches
                    if switch.line.sensor is not None
                ]
            )
            num_fails = 0
            need_manual_attention = False
            # Loop sensors and count failed ones
            for sensor in sensors:
                # If no ICT network
                if self.ict_node is None:
                    (
                        repair_time,
                        line_fail_status,
                    ) = sensor.get_line_fail_status(dt)
                    self.sectioning_time += repair_time
                # If both components have ICT nodes
                elif self.ict_node is not None and sensor.ict_node is not None:
                    # If the ICT nodes are connected to each other
                    if is_connected(
                        node_1=self.ict_node,
                        node_2=sensor.ict_node,
                        network=self.ict_network,
                    ):
                        (
                            repair_time,
                            line_fail_status,
                        ) = sensor.get_line_fail_status(dt)
                        self.sectioning_time += repair_time
                    # If the ICT nodes are not connected to each other
                    else:
                        need_manual_attention = True
                        line_fail_status = sensor.line.failed
                # If no ICT node on sensor
                else:
                    need_manual_attention = True
                    line_fail_status = sensor.line.failed
                num_fails += 1 if line_fail_status else 0
            if need_manual_attention is True:
                self.sectioning_time += self.manual_sectioning_time
            if num_fails > 0:
                section.state = SectionState.DISCONNECTED
                self.sectioning_time += section.get_disconnect_time(dt, self)
                self.failed_sections.append(section)
                self.failed_sections = unique(self.failed_sections)

    def run_control_loop(self, curr_time: Time, dt: Time):
        """
        System control check, determines if components have failed and performes
        the required action

        Parameters
        ----------
        curr_time : Time
            The current time
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        self.sectioning_time = (
            self.sectioning_time - dt
            if self.sectioning_time > Time(0)
            else Time(0)
        )
        self.sectioning_time = max(
            self.sectioning_time,
            self.parent_sectioning_time,
        )
        if (
            self.power_network.connected_line.circuitbreaker.is_open
            and self.sectioning_time <= Time(0)
        ):
            self.check_components = True
        if self.check_components:
            self.check_sensors(curr_time, dt)
            self.check_components = False
        self.check_circuitbreaker(curr_time, dt)

    def check_lines_manually(self, curr_time):
        """
        Loops manually through the sections connected to the controller determining
        which lines have failed. Performs actions according
        to the line status in the respective section.

        If a section was disconnected and no longer includes any failed
        lines, it is connected.

        If a section was connected and now includes a failed line,
        it is disconnected.

        The total sectioning time is summed from each section.

        Parameters
        ----------
        curr_time : Time
            The current time

        Returns
        ----------
        None

        """
        connected_sections = [
            x
            for x in self.power_network.sections
            if x.state == SectionState.CONNECTED
        ]
        disconnected_sections = [
            x
            for x in self.power_network.sections
            if x.state == SectionState.DISCONNECTED
        ]
        # Loop disconnected sections
        for section in disconnected_sections:
            if sum(x.failed for x in section.lines) == 0:
                section.connect_manually()
                if section in self.failed_sections:
                    self.failed_sections.remove(section)
        # Loop connected sections
        for section in connected_sections:
            if sum(x.failed for x in section.lines) > 0:
                section.state = SectionState.DISCONNECTED
                self.failed_sections.append(section)
                self.failed_sections = unique(self.failed_sections)
                self.sectioning_time = self.manual_sectioning_time
                for line in section.lines:
                    line.remaining_outage_time += self.sectioning_time

    def run_manual_control_loop(self, curr_time: Time, dt: Time):
        """
        Manual system control check, determines if components have failed and
        performes the required action.

        Parameters
        ----------
        curr_time : Time
            The current time
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        self.sectioning_time = (
            self.sectioning_time - dt
            if self.sectioning_time > Time(0)
            else Time(0)
        )
        self.sectioning_time = max(
            self.sectioning_time,
            self.parent_sectioning_time,
        )
        if (
            self.power_network.connected_line.circuitbreaker.is_open
            and self.sectioning_time <= Time(0)
        ):
            self.check_components = True
        if self.check_components:
            self.check_lines_manually(curr_time)
            self.check_components = False
        self.check_circuitbreaker_manually(curr_time, dt)

    def set_sectioning_time(self, sectioning_time):
        """
        Sets the sectiom time of the microgrid based on the max value
        of the microgrid sectioning time and the sectioning time set by the
        controller

        Parameters
        ----------
        sectioning_time : Time
            The sectioning time in the microgrid

        Returns
        ----------
        None

        """
        self.sectioning_time = max(
            self.sectioning_time,
            sectioning_time,
        )

    def set_parent_sectioning_time(self, sectioning_time):
        """
        Sets the sectioning time of the parent power network

        Parameters
        ----------
        sectioning_time : Time
            The sectioning time in the microgrid

        Returns
        ----------
        None

        """
        self.parent_sectioning_time = sectioning_time

    def update_fail_status(self, dt: Time):
        """
        Updates the fail status

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        pass

    def update_history(
        self, prev_time: Time, curr_time: Time, save_flag: bool
    ):
        """
        Updates the history variables

        Parameters
        ----------
        prev_time : Time
            The previous time
        curr_time : Time
            Current time
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        if save_flag:
            time = curr_time.get_unit_quantity(curr_time.unit)
            self.history["sectioning_time"][
                time
            ] = self.sectioning_time.get_unit_quantity(curr_time.unit)

    def get_history(self, attribute: str):
        """
        Returns the history variables of an attribute

        Parameters
        ----------
        attribute : str
            Controller attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute

        """
        return self.history[attribute]

    def add_random_instance(self, random_gen):
        """
        Adds global random seed

        Parameters
        ----------
        random_gen : int
            Random number generator

        Returns
        ----------
        None

        """
        pass

    def print_status(self):
        """
        Prints the status

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """

    def reset_status(self, save_flag: bool):
        """
        Resets and sets the status of the class parameters

        Parameters
        ----------
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        self.sectioning_time = Time(0)
        if save_flag:
            self.initialize_history()

    def initialize_history(self):
        """
        Initializes the history variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.history["sectioning_time"] = {}



import numpy as np

from relsad.network.components import Bus, MicrogridMode
from relsad.network.systems import PowerSystem, SubSystem, Transmission
from relsad.Time import Time, TimeStamp, TimeUnit
from relsad.topology.sub_systems import find_backup_lines_between_sub_systems
from relsad.utils import subtract, unique


def find_sub_systems(p_s: PowerSystem, curr_time: Time):
    """
    Function that finds the independent sub systems of the given power system
    and adds them to the sub_systems list of the power system

    Parameters
    ----------
    p_s : PowerSystem
        A PowerSystem object
    curr_time : Time
        Current time

    Returns
    ----------
    None

    """

    p_s.sub_systems = []
    # Will only include connected lines
    active_lines = [line for line in p_s.lines if line.connected]
    used_buses = []
    used_lines = []
    sub_system = SubSystem()

    while not (len(used_buses) + len(used_lines)) == (
        len(p_s.buses) + len(active_lines)
    ):
        for bus in p_s.buses:
            if bus not in unique(used_buses + sub_system.buses):
                if (len(sub_system.buses) + len(sub_system.lines)) == 0:
                    sub_system, used_buses = add_bus(
                        p_s,
                        bus,
                        sub_system,
                        used_buses,
                    )
                    (
                        sub_system,
                        used_buses,
                        used_lines,
                    ) = try_to_add_connected_lines(
                        p_s,
                        bus,
                        sub_system,
                        used_buses,
                        used_lines,
                    )
                    p_s.sub_systems.append(sub_system)
                    p_s.sub_systems = unique(p_s.sub_systems)
                    if (len(used_buses) + len(used_lines)) == (
                        len(p_s.buses) + len(active_lines)
                    ):
                        break
                    sub_system = SubSystem()

    if len(p_s.sub_systems) > 1:
        update_backup_lines_between_sub_systems(p_s, curr_time)


def try_to_add_connected_lines(
    p_s: PowerSystem,
    bus: Bus,
    sub_system: SubSystem,
    used_buses: list,
    used_lines: list,
):
    """
    Add lines to the sub system

    Parameters
    ----------
    p_s : PowerSystem
        A PowerSystem object
    bus : Bus
        A Bus object
    sub_system : SubSystem
        A SubSystem object
    used_buses : list
        List of used Bus elements
    used_lines : list
        List of used Line elements

    Returns
    ----------
    sub_system : SubSystem
        A SubSystem object
    used_buses : list
        List of used Bus elements
    used_lines : list
        List of used Line elements

    """
    for line in subtract(bus.connected_lines, used_lines):
        if line.connected:
            sub_system.add_line(line)
            used_lines.append(line)
            used_lines = unique(used_lines)
            if line.tbus == bus:
                sub_system, used_buses = add_bus(
                    p_s,
                    line.fbus,
                    sub_system,
                    used_buses,
                )
                (
                    sub_system,
                    used_buses,
                    used_lines,
                ) = try_to_add_connected_lines(
                    p_s,
                    line.fbus,
                    sub_system,
                    used_buses,
                    used_lines,
                )
            else:
                sub_system, used_buses = add_bus(
                    p_s,
                    line.tbus,
                    sub_system,
                    used_buses,
                )
                (
                    sub_system,
                    used_buses,
                    used_lines,
                ) = try_to_add_connected_lines(
                    p_s,
                    line.tbus,
                    sub_system,
                    used_buses,
                    used_lines,
                )
    return sub_system, used_buses, used_lines


def add_bus(
    p_s: PowerSystem,
    bus: Bus,
    sub_system: SubSystem,
    used_buses: list,
):
    """
    Add buses to the sub system

    Parameters
    ----------
    p_s : PowerSystem
        A PowerSystem object
    bus : Bus
        A Bus object
    sub_system : SubSystem
        A SubSystem object
    used_buses : list
        List of used Bus elements

    Returns
    ----------
    sub_system : SubSystem
        A SubSystem object
    used_buses : list
        List of used Bus elements

    """
    if bus not in unique(used_buses + sub_system.buses):
        sub_system.add_bus(bus)
        used_buses.append(bus)
        used_buses = unique(used_buses)
        for child_network in p_s.child_network_list:
            if bus in child_network.buses:
                sub_system.add_child_network(child_network)
    return sub_system, used_buses


def update_backup_lines_between_sub_systems(p_s: PowerSystem, curr_time: Time):
    """
    Function that updates the backup lines between the sub systems
    of the power system if they exist and are not failed

    Parameters
    ----------
    p_s : PowerSystem
        A PowerSystem element
    curr_time : Time
        Current time

    Returns
    ----------
    None

    """
    update = False
    for s_1 in p_s.sub_systems:
        for s_2 in p_s.sub_systems:
            if s_1 != s_2:
                external_backup_lines = find_backup_lines_between_sub_systems(
                    s_1, s_2
                )
                for line in external_backup_lines:
                    if (
                        not line.connected
                        and not line.failed
                        and all(
                            [
                                x.parent_network.controller.sectioning_time
                                <= Time(0)
                                for x in line.tbus.connected_lines
                                + line.fbus.connected_lines
                            ]
                        )
                    ):
                        for discon in line.get_disconnectors():
                            if discon.is_open:
                                discon.close()
                        update = True
                        break
            if update:
                break
        if update:
            break
    if update:
        find_sub_systems(p_s, curr_time)


def update_sub_system_slack(p_s: PowerSystem):
    """
    Function that updates the current slack bus of the sub systems of the power system

    Parameters
    ----------
    p_s : PowerSystem
        A PowerSystem element

    Returns
    ----------
    None

    """
    possible_sub_systems = list(p_s.sub_systems)
    for sub_system in possible_sub_systems:
        sub_system.slack = None
        for bus in sub_system.buses:
            bus.is_slack = False
            if set_slack(p_s, sub_system):
                break


def set_slack(p_s: PowerSystem, sub_system: SubSystem):
    """
    Function that sets the slack bus of the power system

    Parameters
    ----------
    p_s : PowerSystem
        A PowerSystem element
    sub_system : SubSystem
        A SubSystem element

    Returns
    ----------
    bool
        Success/Failure of operation

    """
    ## Transmission network slack buses in sub_system
    for bus in sub_system.buses:
        for child_network in p_s.child_network_list:
            if isinstance(child_network, Transmission):
                if bus == child_network.get_trafo_bus():
                    bus.set_slack()
                    sub_system.slack = bus
                    return True
    ## Buses with battery
    if sub_system.slack is None:
        for bus in sub_system.buses:
            if (
                bus.battery is not None and bus.battery.mode is None
            ):  # Battery in Distribution network
                bus.set_slack()
                sub_system.slack = bus
                return True
        for bus in sub_system.buses:
            if (
                bus.battery is not None and bus.battery.mode is not None
            ):  # Battery in Microgrid
                if (
                    bus.battery.mode == MicrogridMode.LIMITED_SUPPORT
                    and bus.battery.remaining_survival_time == Time(0)
                    and not bus.is_slack
                ):
                    bus.battery.start_survival_time()
                bus.set_slack()
                sub_system.slack = bus
                return True
    ## Buses with production
    if sub_system.slack is None:
        for bus in sub_system.buses:
            if bus.prod is not None:
                bus.set_slack()
                sub_system.slack = bus
                return True
    ## Buses with EV_park
    if sub_system.slack is None:
        for bus in sub_system.buses:
            if bus.ev_park is not None and bus.ev_park.v2g_flag is True:
                bus.set_slack()
                sub_system.slack = bus
                return True
    ## Not slack material
    return False


def prepare_system(
    power_system: PowerSystem,
    start_time: TimeStamp,
    stop_time: TimeStamp,
    time_step: Time,
    time_unit: TimeUnit,
):
    """
    Prepares the power system for a simulation by:
     - Creating system sections based on switches
     - Defining the simulation time increments
     - Interpolating load and production data based on
       the time increments

    Parameters
    ----------
    p_s : PowerSystem
        A PowerSystem element
    start_time : TimeStamp
        The start time of the simulation/iteration
    stop_time : TimeStamp
        The stop time of the simulation/iteration
    time_step : Time
        A time step (1 hour, 2 hours, ect.)
    time_unit : TimeUnit
        A time unit (hour, seconds, ect.)

    Returns
    ----------
    time_array:
        Array containing the time increments of the simulation

    """
    # Create power system sections
    power_system.create_sections()

    # Set up time increments based on defined period
    increments = int((stop_time - start_time) / time_step)
    sim_duration = increments * time_step.get_unit_quantity(time_unit)
    time_array = np.arange(
        start=time_step.get_unit_quantity(time_unit),
        stop=sim_duration,
        step=time_step.get_unit_quantity(time_unit),
    )
    time_array_indices = np.arange(increments)

    # Prepare load and production data
    power_system.prepare_load_data(time_array_indices)
    power_system.prepare_prod_data(time_array_indices)

    return time_array


def reset_system(power_system: PowerSystem, save_flag: bool):
    """
    Resets the power system

    Parameters
    ----------
    p_s : PowerSystem
        A PowerSystem element
    save_flag : bool
        Indicates if saving is on or off

    Returns
    ----------
    None

    """
    power_system.reset_energy_shed_variables()
    for network in power_system.child_network_list:
        network.reset_energy_shed_variables()

    for comp in power_system.comp_list:
        comp.reset_status(save_flag)
    power_system.controller.reset_status(save_flag)

    ## Find sub systems
    find_sub_systems(power_system, 0)
    update_sub_system_slack(power_system)



from relsad.network.components import Bus, Line
from relsad.utils import unique


def get_downstream_buses(parent_bus: Bus):
    """
    Function that finds and returns all downstream buses in a radial tree

    Parameters
    ----------
    parent_bus : Bus
        Parent bus of radial tree

    Returns
    -------
    downstream_buses : list
        List of all downstream buses from parent_bus

    """
    if len(parent_bus.nextbus) == 0:
        return [[[parent_bus]]]
    downstream_buses = []
    for nbus in parent_bus.nextbus:
        for downstream_bus in get_downstream_buses(nbus):
            downstream_buses.append([[parent_bus]] + downstream_bus)
    return downstream_buses


def configure_bfs_load_flow_setup(
    bus_list: list,
    line_list: list,
):
    """
    Function that sets up the nested topology array and configures the radial tree according to the slack bus

    Parameters
    ----------
    bus_list : list
        List containing the bus elements
    line_list : list
        List containing the line elements

    Returns
    -------
    topology : list
        Nested topology list
    bus_list : list
        Updated bus list

    """

    ## Find slack bus
    for i, bus in enumerate(bus_list):
        if bus.is_slack:
            slack_bus = bus
            old = bus_list[0]
            bus_list[0] = slack_bus
            bus_list[i] = old
            break

    # Update directions based on slack bus
    # (making slack bus parent of the radial tree)
    update_line_direction_based_on_slack_bus(slack_bus, bus_list, line_list)

    downstream_bus_list = get_downstream_buses(slack_bus)

    topology_bus_list = get_topology_bus_list(downstream_bus_list)

    return topology_bus_list, bus_list


def line_between_buses(
    bus1: Bus,
    bus2: Bus,
    line_list: list,
):
    """
    Returns the line between two buses (bus1 and bus2)
    if it exists, else None

    Parameters
    ----------
    bus1 : Bus
        A Bus element
    bus2 : Bus
        A Bus elememt
    line_list : list
        List containing all the lines

    Returns
    -------
    line : Line
        The line between bus1 and bus2

    """
    for line in line_list:
        if (line.tbus == bus1 and line.fbus == bus2) or (
            line.tbus == bus2 and line.fbus == bus1
        ):
            return line
    return None


def update_line_direction_based_on_slack_bus(
    slack_bus: Bus,
    bus_list: list,
    line_list: list,
):
    """
    Update line directions based on slack bus
    (making slack bus parent of the radial tree)

    Parameters
    ----------
    slack_bus : Bus
        Slack bus
    bus_list : list
        List containing buses
    line_list : list
        List containing lines

    Returns
    -------
    None

    """
    target_buses = [slack_bus]
    used_target_buses = list()
    while target_buses != list():
        new_target_buses = list()
        for target_bus in target_buses:
            if target_bus not in used_target_buses:
                for bus in bus_list:
                    if bus not in used_target_buses and bus != target_bus:
                        line = line_between_buses(target_bus, bus, line_list)
                        if line is not None:
                            if target_bus in bus.nextbus:
                                line.change_direction()
                            new_target_buses.append(bus)
                            new_target_buses = unique(new_target_buses)
                used_target_buses.append(target_bus)
                used_target_buses = unique(used_target_buses)
        target_buses = new_target_buses


def get_topology_bus_list(downstream_bus_list: list):
    """
    Function that constructs a nested topology bus list

    Parameters
    ----------
    downstream_bus_list : list
        Nested list of all downstream buses from parent_bus
        in radial tree

    Returns
    ----------
    topology_bus_list : list
        Nested topology list

    """
    used_buses = list()
    main_downstream_bus_list = downstream_bus_list[0]
    used_buses += main_downstream_bus_list
    topology_bus_list = [main_downstream_bus_list]
    for downstream_child_bus_list in downstream_bus_list[1:]:
        sub_downstream_bus_list = list()
        for bus in downstream_child_bus_list:
            if bus not in used_buses:
                sub_downstream_bus_list.append(bus)
        used_buses += sub_downstream_bus_list
        if sub_downstream_bus_list != list():
            topology_bus_list.append(sub_downstream_bus_list)

    while len(topology_bus_list) > 1:
        last_downstream_bus_list = topology_bus_list[-1]
        top_bus = last_downstream_bus_list[0][0]
        for n, downstream_child_bus_list in enumerate(topology_bus_list[:-1]):
            for k, bus in enumerate(downstream_child_bus_list):
                if top_bus in bus[0].nextbus:
                    topology_bus_list[n][k].append(last_downstream_bus_list)
                    topology_bus_list.remove(last_downstream_bus_list)
                    break

    topology_bus_list = topology_bus_list[0]

    return topology_bus_list


def is_cyclic(network):
    """
    Function that checks if the network contains cycles
    using a depth first search

    Parameters
    ----------
    network : PowerNetwork
        Power network

    Returns
    -------
    cyclic : bool
        Boolean variable stating whether the network contains
        cycles

    """
    cyclic = False

    visited_buses = {bus: False for bus in network.buses}
    stack = []

    for bus in network.buses:

        if visited_buses[bus] is True:
            continue

        stack.append([bus, None])

        while len(stack) > 0:
            curr_bus, parent_bus = stack.pop()
            visited_buses[curr_bus] = True

            for neighbor_bus in curr_bus.get_neighbor_buses():
                if visited_buses[neighbor_bus] is False:
                    stack.append([neighbor_bus, curr_bus])
                    visited_buses[neighbor_bus] = True
                elif neighbor_bus != parent_bus:
                    cyclic = True
                    return cyclic

    return cyclic



from relsad.network.components import Bus, DistributionController, Line
from relsad.utils import unique

from .PowerNetwork import PowerNetwork
from .PowerSystem import PowerSystem


class Distribution(PowerNetwork):

    """
    Class defining a distribution network type

    ...

    Attributes
    ----------
    name : str
        Name of the distribution network
    buses : list
        List containing the buses connected to the distribution network
    ev_parks : list
        List containing the EV parks in the distribution network
    batteries : list
        List containing the batteries in the distribution network
    productions : list
        List containing the generation units in the distribution network
    lines : list
        List containing the lines connected to the distribution network
    sensors : list
        List containing the sensors in the distribution network
    circuitbreaker : list
        List containing the circuit breakers in the distribution network
    disconnectors : list
        List containing the disconnectors in the distribution network
    intelligent_switches : list
        List containing the intelligent switches in the distribution network
    controller : DistributionController
        The controller for the distribution system
    comp_list : list
        List containing the components in the distribution network
    comp_dict : dict
        Dictionary containing the components in the distribution network
    parent_network : PowerNetwork
        The parent network of the distribution network
    power_system : PowerNetwork
        Connects the distribution network to a power system
    child_network_list : list
        List containing connected child networks to the distribution network
    failed_line : Bool
        Flag stating if the distribution network contains a failed line
    p_energy_shed : float
        Shedded active power load in the distribution network
    acc_p_energy_shed : float
        The accumulated shedded active power load in the distribution network
    q_energy_shed : float
        Shedded reactive power load in the distribution network
    acc_q_energy_shed : float
        The accumulated shedded reactive power load in the distribution network
    connected_line : Line
        Connects the distribution network to the transmission network, the line connecting the distribution network to the transmission network
    add_connected_line : Line
        Adds the connected line to the distribution network
    sections : Section
        The sections in the distribution system
    history : dict
        Dictionary containing the history variables of the network
    monte_carlo_history : dict
        Dictionary containing the history variables from the monte carlo simulation



    Methods
    ----------
    add_connected_line(connected_line)
        Sets the line connecting the distribution system to overlying network
    add_bus(bus)
        Adding a bus including elements on the bus (battery, generation unit, EV park) to the distribution network
    add_buses(buses)
        Adding buses to the distribution network
    add_line(line)
        Adding a line including elements on the line (sensor, circuit breaker, disconnector) to the distribution network
    add_lines(lines)
        Adding lines to the distribution network
    get_lines()
        Returns the lines in the distribution network
    reset_slack_bus()
        Resets the slack bus attribute of the buses in the distribution network
    add_child_network(network)
        Adds child network to the distribution network
    get_monte_carlo_history(attribute)
        Returns the specified history variable from the Monte Carlo simulation
    get_history(attribute)
        Returns the specified history variable
    get_system_load()
        Returns the load in the distribution network at the current time in MW and MVar
    reset_energy_shed_variables()
        Resets the energy.shed variables
    """

    ## Visual attributes
    color = "steelblue"

    ## Counter
    counter = 0

    def __init__(self, parent_network: PowerNetwork, connected_line: Line):
        Distribution.counter += 1
        self.name = "dist_network{:d}".format(Distribution.counter)

        # Components
        self.buses = list()
        self.ev_parks = list()
        self.batteries = list()
        self.productions = list()
        self.lines = list()
        self.sensors = list()
        self.circuitbreakers = list()
        self.disconnectors = list()
        self.intelligent_switches = list()
        self.controller = DistributionController(
            name=self.name + "_controller",
            power_network=self,
        )
        self.comp_list = list()
        self.comp_dict = dict()

        # PowerNetwork connections
        self.parent_network = parent_network
        parent_network.add_child_network(self)
        if isinstance(parent_network, PowerSystem):
            self.power_system = parent_network
        else:
            self.power_system = parent_network.parent_network
        self.power_system.controller.add_distribution_controller(
            self.controller
        )
        self.child_network_list = list()
        self.power_system.comp_dict[self.controller.name] = self.controller
        self.power_system.comp_list.append(self.controller)

        self.failed_line = False
        # Load shedding
        self.p_energy_shed = 0
        self.acc_p_energy_shed = 0
        self.q_energy_shed = 0
        self.acc_q_energy_shed = 0

        self.connected_line = None
        if connected_line is not None:
            self.add_connected_line(connected_line)

        # Sectioning
        self.sections = None
        ## History
        self.history: dict = {}
        self.monte_carlo_history: dict = {}

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Distribution(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def add_connected_line(self, connected_line):
        """
        Sets the line connecting the distribution system to overlying network

        Parameters
        ----------
        connected_line : Line
            The line connecting the distribution system to overlaying network

        Returns
        ----------
        None

        """
        # Sets the connected line for the distribution system
        self.connected_line = connected_line

        # Add the components attached to the line
        # to the distribution system:

        # CircuitBreaker
        c_b = connected_line.circuitbreaker
        if self.connected_line.circuitbreaker is None:
            raise Exception(
                "{} connects the distribution network to the "
                "transmission network, it must contain a circuitbreaker".format(
                    connected_line
                )
            )
        self.comp_dict[c_b.name] = c_b
        self.comp_list.append(c_b)

        # Line
        self.add_line(connected_line)

    def add_bus(self, bus: Bus):
        """
        Adding a bus including elements on the bus (battery, generation unit, EV park) to the distribution network

        Parameters
        ----------
        bus : Bus
            A Bus element

        Returns
        ----------
        None

        """
        # Add bus to distribution network
        self.comp_dict[bus.name] = bus
        bus.handle.color = self.color
        bus.color = self.color
        bus.parent_network = self
        self.buses.append(bus)
        self.buses = unique(self.buses)

        # Add components attached to bus to distribution network:

        # EV-Park
        if bus.ev_park is not None:
            self.comp_dict[bus.ev_park.name] = bus.ev_park
            self.comp_list.append(bus.ev_park)
            self.ev_parks.append(bus.ev_park)
            self.ev_parks = unique(self.ev_parks)

        # Battery
        if bus.battery is not None:
            self.comp_dict[bus.battery.name] = bus.battery
            self.comp_list.append(bus.battery)
            self.batteries.append(bus.battery)
            self.batteries = unique(self.batteries)

        # Production
        if bus.prod is not None:
            self.comp_dict[bus.prod.name] = bus.prod
            self.comp_list.append(bus.prod)
            self.productions.append(bus.prod)
            self.productions = unique(self.productions)

        # Add bus to the power system
        self.power_system.add_bus(bus)

    def add_buses(self, buses: list):
        """
        Adding buses to the distribution network bus list

        Parameters
        ----------
        buses : list
            A list of Bus elements in the distribution network

        Returns
        ----------
        None

        """
        for bus in buses:
            self.add_bus(bus)

    def add_line(self, line: Line):
        """
        Adding a line and the components connected to the line to the distribution network

        Parameters
        ----------
        line : Line
            A Line element

        Returns
        ----------
        None

        """

        # Add line to distribution network
        line.handle.color = self.color
        line.color = self.color
        self.comp_dict[line.name] = line
        self.lines.append(line)
        self.lines = unique(self.lines)

        # Add components attached to line to distribution network:

        # Sensor
        if line.sensor:
            self.comp_dict[line.sensor.name] = line.sensor
            self.comp_list.append(line.sensor)
            self.sensors.append(line.sensor)
            self.sensors = unique(self.sensors)
            self.controller.sensors.append(line.sensor)
            self.controller.sensors = unique(self.controller.sensors)

        # Disconnector
        for discon in line.disconnectors:
            self.comp_dict[discon.name] = discon
            self.comp_list.append(discon)
            self.disconnectors.append(discon)
            self.disconnectors = unique(self.disconnectors)

            # Intelligent switch
            if discon.intelligent_switch:
                self.comp_dict[
                    discon.intelligent_switch.name
                ] = discon.intelligent_switch
                self.comp_list.append(discon.intelligent_switch)
                self.intelligent_switches.append(discon.intelligent_switch)
                self.intelligent_switches = unique(self.intelligent_switches)

        # Set distribution network as parent network
        line.add_parent_network(self)

        # Add line to power system
        self.power_system.add_line(line)

    def add_lines(self, lines: list):
        """
        Adding lines to distribution network line list

        Parameters
        ----------
        lines : list
            A list of Line elements in the distribution network

        Returns
        ----------
        None

        """
        for line in lines:
            self.add_line(line)

    def get_lines(self):
        """
        Returns the lines in the distribution network

        Parameters
        ----------
        None

        Returns
        ----------
        lines : list
            List of Line elements

        """
        return self.lines

    def reset_slack_bus(self):
        """
        Resets the slack bus attribute of the buses in the distribution network

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        for bus in self.buses:
            bus.is_slack = False

    def add_child_network(self, network):
        """
        Adds child network to the distribution network

        Parameters
        ----------
        network : PowerNetwork
            The child network of the distribution network

        Returns
        ----------
        None

        """
        self.child_network_list.append(network)
        self.parent_network.add_child_network(network)

    def get_monte_carlo_history(self, attribute):
        """
        Returns the specified history variable from the Monte Carlo simulation

        Parameters
        ----------
        attribute : str
            Distribution network attribute

        Returns
        ----------
        monte_carlo_history[attribute] : dict
            Returns the history variables of an attribute from the Monte Carlo simulation

        """
        return self.monte_carlo_history[attribute]

    def get_history(self, attribute):
        """
        Returns the specified history variable

        Parameters
        ----------
        attribute : str
            Distribution network attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute

        """
        return self.history[attribute]

    def get_system_load(self):
        """
        Returns the system load in the distribution network at the current time in MW and MVar

        Parameters
        ----------
        None

        Returns
        ----------
        pload : float
            The active power load in the distribution network
        qload : float
            The reactive power load in the distribution network

        """
        pload, qload = 0, 0
        for bus in self.buses:
            p, q = bus.get_load()
            pload += p
            qload += q
        return pload, qload

    def reset_energy_shed_variables(self):
        """
        Resets the energy.shed variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.p_energy_shed = 0
        self.acc_p_energy_shed = 0
        self.q_energy_shed = 0
        self.acc_q_energy_shed = 0



from enum import Enum

import matplotlib.lines as mlines
import numpy as np

from relsad.Time import Time, TimeUnit
from relsad.utils import convert_yearly_fail_rate, random_choice

from .Component import Component
from .ICTNode import ICTNode
from .Line import Line


class SensorState(Enum):
    """
    Line sensor state

    Attributes
    ----------
    OK : int
        The sensor is up and running
    FAILED : int
        The sensor has failed
    REPAIR : int
        The sensor is being repaired
    """

    OK = 1
    FAILED = 2
    REPAIR = 3


class Sensor(Component):

    """
    Class defining line sensors
    ...

    Attributes
    ----------
    name : string
        Name of the sensor
    line : Line
        The line the sensor is connected to
    coordinate : list
        The coordinate of the sensor
    ict_node : ICTNode
        The ICT node connected to the sensor
    fail_rate_per_year : float
        The failure rate per year for the sensor
    p_fail_repair_new_signal : float
        The probability of a new signal can be sent
    p_fail_repair_reboot : float
        The probability that a reboot of the sensor works
    new_signal_time : Time
        The time it takes to send a new signal
    reboot_time : Time
        The time it takes to reboot the sensor
    manual_repair_time : Time
        The time it takes to manually repair the sensor
    state : SensorState
        Which state the sensor is in
    remaining_repair_time : Time
        The remaining repair time of the intelligent switch
    history : dict
        Dictonary attribute that stores the historic variables
    monte_carlo_history : dict
        Dictonary attribute that stores the historic variables from the Monte Carlo simulation

    Methods
    ----------
    fail()
        Sets the sensor state to FAILED
    not_fail()
        Sets the sensor state to OK
    draw_fail_status(dt)
        Draws the state of the sensor for a given time step
    draw_status(prob)
        Sets the state of the sensor based on the probability of the state being FAILED
    repair(dt)
        Returns the repair time and fail status of the sensor
    get_line_fail_status(dt)
        Returns the repair time and fail status of the line
    get_section()
        Returns the line section
    update_fail_status(dt)
        Updates the fail status of the sensor
        If the state of the sensor is REPAIR, the remaining repair time is set
        If the state of the sensor is OK, the state of the sensor is drawn
    update_history(prev_time, curr_time, save_flag)
        Updates the history variables
    get_history(attribute)
        Returns the history variables of an attribute
    add_random_instance(random_gen)
        Adds global random instance
    print_status()
        Prints the status
    reset_status(save_flag)
        Resets the status of the sensor
    initialize_history()
        Initializes the history variables
    """

    ## Visual attributes
    color = "rosybrown"
    marker = "s"
    size = 2**2
    handle = mlines.Line2D(
        [],
        [],
        marker=marker,
        markeredgewidth=1,
        markersize=size,
        color=color,
        linestyle="None",
    )

    ## Random instance
    ps_random: np.random.Generator = None

    def __init__(
        self,
        name: str,
        line: Line,
        ict_node: ICTNode = None,
        fail_rate_per_year: float = 0.023,
        p_fail_repair_new_signal: float = 1 - 0.95,
        p_fail_repair_reboot: float = 1 - 0.9,
        new_signal_time: Time = Time(2, TimeUnit.SECOND),
        reboot_time: Time = Time(5, TimeUnit.MINUTE),
        manual_repair_time: Time = Time(2, TimeUnit.HOUR),
        state: SensorState = SensorState.OK,
    ):

        # Verify input
        if line is None:
            raise Exception("Sensor must be connected to a Line")
        if line.parent_network is not None:
            raise Exception(
                "Sensor must be created before the line is connected to a network"
            )
        if fail_rate_per_year < 0:
            raise Exception("The failure rate per year must be positive")
        if p_fail_repair_new_signal < 0 or p_fail_repair_new_signal > 1:
            raise Exception("p_fail_repair_new_signal must be between 0 and 1")
        if p_fail_repair_reboot < 0 or p_fail_repair_reboot > 1:
            raise Exception("p_fail_repair_reboot must be between 0 and 1")

        self.name = name
        self.line = line
        line.sensor = self

        # Define coordinate
        x1, y1 = line.fbus.coordinate
        x2, y2 = line.tbus.coordinate

        x = (x1 + x2) / 2
        y = (y1 + y2) / 2

        self.coordinate = [x, y]
        self.ict_node = ict_node
        if ict_node is not None:
            ict_node.coordinate = self.coordinate
        self.fail_rate_per_year = fail_rate_per_year
        self.p_fail_repair_new_signal = p_fail_repair_new_signal
        self.p_fail_repair_reboot = p_fail_repair_reboot
        self.new_signal_time = new_signal_time
        self.reboot_time = reboot_time
        self.remaining_repair_time = Time(0)
        self.manual_repair_time = manual_repair_time
        self.state = state

        ## History
        self.history = {}
        self.monte_carlo_history = {}
        self.initialize_history()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Sensor(name={self.name})"

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name and isinstance(other, Sensor)
        else:
            return False

    def __hash__(self):
        return hash(self.name)

    def fail(self):
        """
        Sets the sensor state to FAILED

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.state = SensorState.FAILED

    def not_fail(self):
        """
        Sets the sensor state to OK

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.state = SensorState.OK

    def draw_fail_status(self, dt: Time):
        """
        Draws the state of the sensor for a given time step

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        p_not_fail = convert_yearly_fail_rate(self.fail_rate_per_year, dt)
        self.draw_status(p_not_fail)

    def draw_status(self, prob):
        """
        Sets the state of the sensor based on the probability of the state being FAILED

        Parameters
        ----------
        prob : float
            The probability that the sensor state is FAILED

        Returns
        ----------
        None

        """
        if random_choice(self.ps_random, prob):
            self.fail()
        else:
            self.not_fail()

    def repair(self, dt: Time):
        """
        Returns the repair time and fail status of the sensor

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        repair_time : Time
            The repair time of the sensor
        fail_status : bool
            The fail status of the sensor
        """
        repair_time = self.new_signal_time
        fail_status = None
        self.draw_status(self.p_fail_repair_new_signal)
        if self.state == SensorState.OK:
            fail_status = self.line.failed
            return repair_time, fail_status
        elif self.state == SensorState.FAILED:
            repair_time += self.reboot_time
            self.draw_status(self.p_fail_repair_reboot)
            if self.state == SensorState.OK:
                fail_status = self.line.failed
                return repair_time, fail_status
            elif self.state == SensorState.FAILED:
                self.remaining_repair_time = self.manual_repair_time
                self.state = SensorState.REPAIR
                fail_status = True
                return repair_time, fail_status

    def get_line_fail_status(self, dt: Time):
        """
        Returns the repair time and fail status of the line

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        repair_time : Time
            The repair time of the line
        fail_status : bool
            The fail status of the line

        """
        repair_time = None
        fail_status = None
        if self.state == SensorState.REPAIR:
            repair_time = Time(0)
            fail_status = True
            return repair_time, fail_status
        else:
            if self.state == SensorState.OK:
                repair_time = Time(0)
                fail_status = self.line.failed
                return repair_time, fail_status
            elif self.state == SensorState.FAILED:
                repair_time, fail_status = self.repair(dt)
                return repair_time, fail_status

    def get_section(self):
        """
        Returns the line section

        Parameters
        ----------
        None

        Returns
        ----------
        line_section : Section
            The parent section of the line

        """
        line_section = self.line.section
        return line_section

    def update_fail_status(self, dt: Time):
        """
        Updates the fail status of the sensor
        If the state of the sensor is REPAIR, the remaining repair time is set
        If the state of the sensor is OK, the state of the sensor is drawn

        Parameters
        ----------
        dt : Time
            The current time step

        Returns
        ----------
        None

        """
        if self.state == SensorState.REPAIR:
            self.remaining_repair_time -= dt
            if self.remaining_repair_time <= Time(0):
                self.not_fail()
        elif self.state == SensorState.OK:
            self.draw_fail_status(dt)

    def update_history(self, prev_time, curr_time, save_flag: bool):
        """
        Updates the history variables

        Parameters
        ----------
        prev_time : Time
            The previous time
        curr_time : Time
            The current time
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        if save_flag:
            time = curr_time.get_unit_quantity(curr_time.unit)
            self.history["remaining_repair_time"][
                time
            ] = self.remaining_repair_time.get_unit_quantity(curr_time.unit)
            self.history["state"][time] = self.state.value

    def get_history(self, attribute: str):
        """
        Returns the history variables of an attribute

        Parameters
        ----------
        attribute : str
            System attribute

        Returns
        ----------
        history[attribute] : dict
            Returns the history variables of an attribute
        """
        return self.history[attribute]

    def add_random_instance(self, random_gen):
        """
        Adds global random instance

        Parameters
        ----------
        random_gen : np.random.default_rng()
            Random number generator

        Returns
        ----------
        None

        """
        self.ps_random = random_gen

    def print_status(self):
        """
        Prints the status

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        pass

    def reset_status(self, save_flag: bool):
        """
        Resets the status of the sensor

        Parameters
        ----------
        save_flag : bool
            Indicates if saving is on or off

        Returns
        ----------
        None

        """
        self.state = SensorState.OK
        self.remaining_repair_time = Time(0)
        if save_flag:
            self.initialize_history()

    def initialize_history(self):
        """
        Initializes the history variables

        Parameters
        ----------
        None

        Returns
        ----------
        None

        """
        self.history["remaining_repair_time"] = {}
        self.history["state"] = {}



"""
This module contains array utilities
"""

import numpy as np


def interpolate(array: np.ndarray, time_indices: np.ndarray):
    """
    Returns the array interpolated to match the time indices

    arameters
    ----------
    array : np.ndarray
        Array to interpolate
    time_indices : np.ndarray
        Array of required time indices in need of array values

    Returns
    -------
    interpolated_array : np.ndarray
        The array interpolated to match the time indices

    """
    x = np.linspace(0, array.size - 1, time_indices.size)
    interpolated_array = np.interp(x, np.arange(array.size), array)
    return interpolated_array



from relsad.examples.tutorial.system_components import *

from relsad.network.systems import (
    PowerSystem,
    Transmission,
    Distribution,
)

from relsad.load.bus import CostFunction

from relsad.visualization.plotting import plot_topology

import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


load_household = np.ones(365 * 24) * 0.05  # MW

household = CostFunction(
    A=8.8,
    B=14.7,
)

B2.add_load_data(
    pload_data=load_household,
    cost_function=household,
)

B3.add_load_data(
    pload_data=load_household,
    cost_function=household,
)


ps = PowerSystem(controller=C1)

tn = Transmission(
    parent_network=ps,
    trafo_bus=B1,
)

dn = Distribution(
    parent_network=tn,
    connected_line=L1,
)
dn.add_buses([B2, B3, B4, B5, B6])
dn.add_lines([L2, L3, L4, L5, L6])

fig = plot_topology(
    buses=ps.buses,
    lines=ps.lines,
    bus_text=True,
    line_text=True,
)

fig.savefig(
    "test_network.png",
    dpi=600,
)
os.remove("test_network.png")



from .bfs import (
    accumulate_load,
    calc_bus_voltage_sensitivity_single_phase,
    configure_bfs_load_flow_setup,
    get_load,
    run_bfs_load_flow,
    update_voltage,
)

__all__ = []
for v in dir():
    if not v.startswith("__") and v != "relsad":
        __all__.append(v)



import itertools

from relsad.network.systems import SubSystem
from relsad.utils import intersection, unique


def find_backup_lines_between_sub_systems(
    sub_system1: SubSystem,
    sub_system2: SubSystem,
):
    """
    Finds connections between sub systems (sub_system1 and sub_system2)

    Parameters
    ----------
    sub_system1 : SubSystem
        A SubSystem element
    sub_system2 : SubSystem
        A SubSystem element

    Returns
    -------
    backup_lines : list
        List of backup lines connecting sub systems

    """

    external_backup_lines1 = find_external_backup_lines(sub_system1)
    external_backup_lines2 = find_external_backup_lines(sub_system2)
    # Extract backup lines
    backup_lines = intersection(
        external_backup_lines1,
        external_backup_lines2,
    )
    return backup_lines


def find_external_backup_lines(sub_system: SubSystem):
    """
    Finds lines connected to sub system buses that are connected to external sub systems and returns a list of external backup lines

    Parameters
    ----------
    sub_system : SubSystem
        A SubSystem element

    Returns
    -------
    external_backup_lines : list
        List containing external backup lines

    """
    external_backup_lines = list()
    for bus in sub_system.buses:
        for line in bus.connected_lines:
            if line not in sub_system.lines and line.is_backup:
                external_backup_lines.append(line)
                external_backup_lines = unique(external_backup_lines)
    return external_backup_lines



import numpy as np


class Table:

    """
    Utility class for tables in relsad

    ...

    Attributes
    ----------
    x : np.ndarray
        Table indices
    y : np.ndarray
        Table values


    Methods
    ----------
    get_value(value)
        Returns table value based on index

    """

    def __init__(
        self,
        x: np.ndarray,
        y: np.ndarray,
    ):
        # Check duplicate indices
        if len(np.unique(x)) != len(x):
            raise Exception("Tables indices are not unique")
        self.x = x
        self.y = y

    def get_value(self, index):
        """
        Returns table value based on index

        Parameters
        ----------
        index : int
            Table index

        Returns
        ----------
        value : float
            Table value based on index

        """
        if index not in self.x:
            raise Exception("Index {:d} not in table".format(index))
        value = self.y[self.x == index][0]
        return value



import os

import numpy as np

from relsad.network.components import (
    Battery,
    Bus,
    CircuitBreaker,
    Disconnector,
    EVPark,
    IntelligentSwitch,
    Line,
    MainController,
    ManualMainController,
    MicrogridMode,
    Production,
    Sensor,
)
from relsad.network.systems import (
    Distribution,
    Microgrid,
    PowerSystem,
    Transmission,
)
from relsad.StatDist import NormalParameters, StatDist, StatDistType
from relsad.Table import Table
from relsad.Time import Time, TimeUnit
from relsad.visualization.plotting import plot_topology


def initialize_network(
    fail_rate_trafo: float = 0.007,
    fail_rate_line: float = 0.7,
    fail_rate_intelligent_switch: float = 1000,
    fail_rate_hardware: float = 0.2,
    fail_rate_software: float = 12,
    fail_rate_sensor: float = 0.023,
    p_fail_repair_new_signal: float = 1 - 0.95,
    p_fail_repair_reboot: float = 1 - 0.9,
    include_microgrid: bool = True,
    include_production: bool = True,
    include_ICT: bool = True,
    include_ev: bool = True,
    v2g_flag: bool = True,
    include_backup: bool = True,
    line_repair_time_stat_dist=StatDist(
        stat_dist_type=StatDistType.TRUNCNORMAL,
        parameters=NormalParameters(
            loc=1.25,
            scale=1,
            min_val=0.5,
            max_val=2,
        ),
    ),
):
    def num_ev_table_func(
        n_customers,
        ev_percentage=0.47,
        daily_charge_frac=0.61,
    ):
        return Table(
            x=np.array(
                [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    9,
                    10,
                    11,
                    12,
                    13,
                    14,
                    15,
                    16,
                    17,
                    18,
                    19,
                    20,
                    21,
                    22,
                    23,
                ],
            ),
            y=np.array(
                [
                    0.52,
                    0.52,
                    0.52,
                    0.52,
                    0.52,
                    0.52,
                    0.52,
                    0.08,
                    0.08,
                    0.18,
                    0.18,
                    0.18,
                    0.18,
                    0.18,
                    0.18,
                    0.18,
                    0.28,
                    0.28,
                    0.28,
                    0.28,
                    0.42,
                    0.42,
                    0.42,
                    0.42,
                ]
            )
            * n_customers
            * ev_percentage
            * daily_charge_frac,
        )

    if include_ICT:
        C1 = MainController(
            "C1",
            hardware_fail_rate_per_year=fail_rate_hardware,
            software_fail_rate_per_year=fail_rate_software,
            p_fail_repair_new_signal=p_fail_repair_new_signal,
            p_fail_repair_reboot=p_fail_repair_reboot,
        )
    else:
        C1 = ManualMainController(
            name="C1",
            sectioning_time=Time(0, TimeUnit.HOUR),
        )

    ps = PowerSystem(C1)

    ## Transmission network
    B1 = Bus("B1", n_customers=0, coordinate=[0, 0], fail_rate_per_year=0)

    B2 = Bus(
        "B2",
        n_customers=1,
        coordinate=[1, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B3 = Bus(
        "B3",
        n_customers=26,
        coordinate=[2, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B4 = Bus(
        "B4",
        n_customers=26,
        coordinate=[3, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B5 = Bus(
        "B5",
        n_customers=1,
        coordinate=[4, 0],
        fail_rate_per_year=fail_rate_trafo,
    )
    B6 = Bus(
        "B6",
        n_customers=26,
        coordinate=[5, 0],
        fail_rate_per_year=fail_rate_trafo,
    )

    B7 = Bus(
        "B7",
        n_customers=26,
        coordinate=[3, 1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B8 = Bus(
        "B8",
        n_customers=1,
        coordinate=[4, 1],
        fail_rate_per_year=fail_rate_trafo,
    )

    B9 = Bus(
        "B9",
        n_customers=26,
        coordinate=[5, 1],
        fail_rate_per_year=fail_rate_trafo,
    )
    B10 = Bus(
        "B10",
        n_customers=1,
        coordinate=[6, 1],
        fail_rate_per_year=fail_rate_trafo,
    )

    L1 = Line(
        "L1",
        B1,
        B2,
        r=0.0922,
        x=0.0470,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L2 = Line(
        "L2",
        B2,
        B3,
        r=0.4930,
        x=0.2511,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L3 = Line(
        "L3",
        B3,
        B4,
        r=0.3660,
        x=0.1864,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L4 = Line(
        "L4",
        B4,
        B5,
        r=0.3811,
        x=0.1941,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L5 = Line(
        "L5",
        B5,
        B6,
        r=0.8190,
        x=0.7070,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )

    L6 = Line(
        "L6",
        B3,
        B7,
        r=0.8190,
        x=0.7070,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L7 = Line(
        "L7",
        B7,
        B8,
        r=0.8190,
        x=0.7070,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )

    L8 = Line(
        "L8",
        B4,
        B9,
        r=0.8190,
        x=0.7070,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )
    L9 = Line(
        "L9",
        B9,
        B10,
        r=0.8190,
        x=0.7070,
        repair_time_dist=line_repair_time_stat_dist,
        fail_rate_density_per_year=fail_rate_line,
    )

    CircuitBreaker("E1", L1)

    DL1a = Disconnector("L1a", L1, B1)
    DL1b = Disconnector("L1b", L1, B2)
    DL2a = Disconnector("L2a", L2, B2)
    DL2b = Disconnector("L2b", L2, B3)
    DL3a = Disconnector("L3a", L3, B3)
    DL3b = Disconnector("L3b", L3, B4)
    DL4a = Disconnector("L4a", L4, B4)
    DL4b = Disconnector("L4b", L4, B5)
    DL5a = Disconnector("L5a", L5, B5)
    DL5b = Disconnector("L5b", L5, B6)
    DL6a = Disconnector("L6a", L6, B3)
    DL6b = Disconnector("L6b", L6, B7)
    DL7a = Disconnector("L7a", L7, B7)
    DL7b = Disconnector("L7b", L7, B8)
    DL8a = Disconnector("L8a", L8, B4)
    DL8b = Disconnector("L8b", L8, B9)
    DL9a = Disconnector("L9a", L9, B9)
    DL9b = Disconnector("L9b", L9, B10)

    if include_backup:
        L10 = Line(
            "L10",
            B4,
            B8,
            r=0.7114,
            x=0.2351,
            repair_time_dist=line_repair_time_stat_dist,
            fail_rate_density_per_year=fail_rate_line,
            capacity=6,
        )
        DL10a = Disconnector("L10a", L10, B4)
        DL10b = Disconnector("L10b", L10, B8)

        L10.set_backup()

    if include_microgrid:
        M1 = Bus(
            "M1",
            n_customers=0,
            coordinate=[6, 0],
            fail_rate_per_year=fail_rate_trafo,
        )
        M2 = Bus(
            "M2",
            n_customers=0,
            coordinate=[7, 0],
            fail_rate_per_year=fail_rate_trafo,
        )
        M3 = Bus(
            "M3",
            n_customers=40,
            coordinate=[7, -1],
            fail_rate_per_year=fail_rate_trafo,
        )

        Battery("Bat1", M1)
        Production("P1", M2)

        ML1 = Line(
            "ML1",
            M1,
            M2,
            r=0.057526629463617,
            x=0.029324854498807,
            repair_time_dist=line_repair_time_stat_dist,
            fail_rate_density_per_year=fail_rate_line,
        )
        ML2 = Line(
            "ML2",
            M1,
            M3,
            r=0.057526629463617,
            x=0.029324854498807,
            repair_time_dist=line_repair_time_stat_dist,
            fail_rate_density_per_year=fail_rate_line,
        )

        L11 = Line(
            "L11",
            B6,
            M1,
            r=0.057526629463617,
            x=0.029324854498807,
            repair_time_dist=line_repair_time_stat_dist,
            fail_rate_density_per_year=fail_rate_line,
        )

        CircuitBreaker("E2", L11)

        DL11a = Disconnector("L11a", L11, B6)
        DL11b = Disconnector("L11b", L11, M1)
        DML1a = Disconnector("ML1a", ML1, M1)
        DML1b = Disconnector("ML1b", ML1, M2)
        DML2a = Disconnector("ML2a", ML2, M1)
        DML2b = Disconnector("ML2b", ML2, M3)

    if include_ICT:

        Sensor(
            "SL1",
            L1,
            fail_rate_per_year=fail_rate_sensor,
            p_fail_repair_new_signal=p_fail_repair_new_signal,
            p_fail_repair_reboot=p_fail_repair_reboot,
        )
        Sensor(
            "SL2",
            L2,
            fail_rate_per_year=fail_rate_sensor,
            p_fail_repair_new_signal=p_fail_repair_new_signal,
            p_fail_repair_reboot=p_fail_repair_reboot,
        )
        Sensor(
            "SL3",
            L3,
            fail_rate_per_year=fail_rate_sensor,
            p_fail_repair_new_signal=p_fail_repair_new_signal,
            p_fail_repair_reboot=p_fail_repair_reboot,
        )
        Sensor(
            "SL4",
            L4,
            fail_rate_per_year=fail_rate_sensor,
            p_fail_repair_new_signal=p_fail_repair_new_signal,
            p_fail_repair_reboot=p_fail_repair_reboot,
        )
        Sensor(
            "SL5",
            L5,
            fail_rate_per_year=fail_rate_sensor,
            p_fail_repair_new_signal=p_fail_repair_new_signal,
            p_fail_repair_reboot=p_fail_repair_reboot,
        )
        Sensor(
            "SL6",
            L6,
            fail_rate_per_year=fail_rate_sensor,
            p_fail_repair_new_signal=p_fail_repair_new_signal,
            p_fail_repair_reboot=p_fail_repair_reboot,
        )
        Sensor(
            "SL7",
            L7,
            fail_rate_per_year=fail_rate_sensor,
            p_fail_repair_new_signal=p_fail_repair_new_signal,
            p_fail_repair_reboot=p_fail_repair_reboot,
        )
        Sensor(
            "SL8",
            L8,
            fail_rate_per_year=fail_rate_sensor,
            p_fail_repair_new_signal=p_fail_repair_new_signal,
            p_fail_repair_reboot=p_fail_repair_reboot,
        )
        Sensor(
            "SL9",
            L9,
            fail_rate_per_year=fail_rate_sensor,
            p_fail_repair_new_signal=p_fail_repair_new_signal,
            p_fail_repair_reboot=p_fail_repair_reboot,
        )

        if include_backup:
            Sensor(
                "SL10",
                L10,
                fail_rate_per_year=fail_rate_sensor,
                p_fail_repair_new_signal=p_fail_repair_new_signal,
                p_fail_repair_reboot=p_fail_repair_reboot,
            )

        IntelligentSwitch(
            "RL1a", DL1a, fail_rate_per_year=fail_rate_intelligent_switch
        )
        IntelligentSwitch(
            "RL1b", DL1b, fail_rate_per_year=fail_rate_intelligent_switch
        )
        IntelligentSwitch(
            "RL2a", DL2a, fail_rate_per_year=fail_rate_intelligent_switch
        )
        IntelligentSwitch(
            "RL2b", DL2b, fail_rate_per_year=fail_rate_intelligent_switch
        )
        IntelligentSwitch(
            "RL3a", DL3a, fail_rate_per_year=fail_rate_intelligent_switch
        )
        IntelligentSwitch(
            "RL3b", DL3b, fail_rate_per_year=fail_rate_intelligent_switch
        )
        IntelligentSwitch(
            "RL4a", DL4a, fail_rate_per_year=fail_rate_intelligent_switch
        )
        IntelligentSwitch(
            "RL4b", DL4b, fail_rate_per_year=fail_rate_intelligent_switch
        )
        IntelligentSwitch(
            "RL5a", DL5a, fail_rate_per_year=fail_rate_intelligent_switch
        )
        IntelligentSwitch(
            "RL5b", DL5b, fail_rate_per_year=fail_rate_intelligent_switch
        )
        IntelligentSwitch(
            "RL6a", DL6a, fail_rate_per_year=fail_rate_intelligent_switch
        )
        IntelligentSwitch(
            "RL6b", DL6b, fail_rate_per_year=fail_rate_intelligent_switch
        )
        IntelligentSwitch(
            "RL7a", DL7a, fail_rate_per_year=fail_rate_intelligent_switch
        )
        IntelligentSwitch(
            "RL7b", DL7b, fail_rate_per_year=fail_rate_intelligent_switch
        )
        IntelligentSwitch(
            "RL8a", DL8a, fail_rate_per_year=fail_rate_intelligent_switch
        )
        IntelligentSwitch(
            "RL8b", DL8b, fail_rate_per_year=fail_rate_intelligent_switch
        )
        IntelligentSwitch(
            "RL9a", DL9a, fail_rate_per_year=fail_rate_intelligent_switch
        )
        IntelligentSwitch(
            "RL9b", DL9b, fail_rate_per_year=fail_rate_intelligent_switch
        )

        if include_backup:
            IntelligentSwitch(
                "RL10a", DL10a, fail_rate_per_year=fail_rate_intelligent_switch
            )
            IntelligentSwitch(
                "RL10b", DL10b, fail_rate_per_year=fail_rate_intelligent_switch
            )

        if include_microgrid:

            Sensor(
                "SL11",
                L11,
                fail_rate_per_year=fail_rate_sensor,
                p_fail_repair_new_signal=p_fail_repair_new_signal,
                p_fail_repair_reboot=p_fail_repair_reboot,
            )

            Sensor(
                "SML1",
                ML1,
                fail_rate_per_year=fail_rate_sensor,
                p_fail_repair_new_signal=p_fail_repair_new_signal,
                p_fail_repair_reboot=p_fail_repair_reboot,
            )
            Sensor(
                "SML2",
                ML2,
                fail_rate_per_year=fail_rate_sensor,
                p_fail_repair_new_signal=p_fail_repair_new_signal,
                p_fail_repair_reboot=p_fail_repair_reboot,
            )

            IntelligentSwitch(
                "RL11a", DL11a, fail_rate_per_year=fail_rate_intelligent_switch
            )
            IntelligentSwitch(
                "RL11b", DL11b, fail_rate_per_year=fail_rate_intelligent_switch
            )

            IntelligentSwitch(
                "RML1a", DML1a, fail_rate_per_year=fail_rate_intelligent_switch
            )
            IntelligentSwitch(
                "RML1b", DML1b, fail_rate_per_year=fail_rate_intelligent_switch
            )
            IntelligentSwitch(
                "RML2a", DML2a, fail_rate_per_year=fail_rate_intelligent_switch
            )
            IntelligentSwitch(
                "RML2b", DML2b, fail_rate_per_year=fail_rate_intelligent_switch
            )

    if include_ev:

        EVPark(
            name="EVB3",
            bus=B3,
            num_ev_dist=num_ev_table_func(B3.n_customers),
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EVB4",
            bus=B4,
            num_ev_dist=num_ev_table_func(B4.n_customers),
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EVB6",
            bus=B6,
            num_ev_dist=num_ev_table_func(B6.n_customers),
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EVB7",
            bus=B7,
            num_ev_dist=num_ev_table_func(B7.n_customers),
            v2g_flag=v2g_flag,
        )
        EVPark(
            name="EVB9",
            bus=B9,
            num_ev_dist=num_ev_table_func(B9.n_customers),
            v2g_flag=v2g_flag,
        )

    tn = Transmission(ps, B1)

    dn = Distribution(tn, L1)

    dn.add_buses([B1, B2, B3, B4, B5, B6, B7, B8, B9, B10])
    if include_backup:
        dn.add_lines([L2, L3, L4, L5, L6, L7, L8, L9, L10])
    else:
        dn.add_lines([L2, L3, L4, L5, L6, L7, L8, L9])

    if include_microgrid:
        m = Microgrid(dn, L11, mode=MicrogridMode.SURVIVAL)

        m.add_buses([M1, M2, M3])
        m.add_lines([ML1, ML2])

    return ps


if __name__ == "__main__":
    ps = initialize_network()
    fig = plot_topology(ps.buses, ps.lines)

    fig.savefig(
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "TEST10.pdf",
        )
    )



from relsad.examples.tutorial.system_components import *

from relsad.network.systems import (
    PowerSystem,
    Transmission,
    Distribution,
)

from relsad.network.systems import Microgrid

from relsad.network.components import MicrogridMode

ps = PowerSystem(controller=C1)

tn = Transmission(
    parent_network=ps,
    trafo_bus=B1,
)

dn = Distribution(
    parent_network=tn,
    connected_line=L1,
)
dn.add_buses([B2, B3, B4, B5, B6])
dn.add_lines([L2, L3, L4, L5, L6])

# Buses:

M1 = Bus(
    name="M1",
    n_customers=1,
    coordinate=[-1, -2],
)

M2 = Bus(
    name="M2",
    n_customers=1,
    coordinate=[-2, -3],
)

M3 = Bus(
    name="M3",
    n_customers=1,
    coordinate=[-1, -3],
)

# Lines:

ML1 = Line(
    name="ML1",
    fbus=B2,
    tbus=M1,
    r=0.5,
    x=0.5,
    fail_rate_density_per_year=fail_rate_line,
    repair_time_dist=line_stat_repair_time_dist,
)

ML2 = Line(
    name="ML2",
    fbus=M1,
    tbus=M2,
    r=0.5,
    x=0.5,
    fail_rate_density_per_year=fail_rate_line,
    repair_time_dist=line_stat_repair_time_dist,
)

ML3 = Line(
    name="ML3",
    fbus=M1,
    tbus=M3,
    r=0.5,
    x=0.5,
    fail_rate_density_per_year=fail_rate_line,
    repair_time_dist=line_stat_repair_time_dist,
)

# Circuit breaker:

E2 = CircuitBreaker(name="E2", line=ML1)

# Disconnectors:

DML1a = Disconnector(
    name="ML1a",
    line=ML1,
    bus=B2,
)
DML1b = Disconnector(
    name="ML1b",
    line=ML1,
    bus=M1,
)
DML2a = Disconnector(
    name="ML2a",
    line=ML2,
    bus=M1,
)
DML2b = Disconnector(
    name="ML2b",
    line=ML2,
    bus=M2,
)
DML3a = Disconnector(
    name="ML3a",
    line=ML3,
    bus=M1,
)
DML4b = Disconnector(
    name="ML4b",
    line=ML3,
    bus=M3,
)

m = Microgrid(
    distribution_network=dn,
    connected_line=ML1,
    mode=MicrogridMode.FULL_SUPPORT,
)
m.add_buses([M1, M2, M3])
m.add_lines([ML2, ML3])
